from http import HTTPStatus
from unittest.mock import ANY

from django.conf import settings
from django.test import Client
from django.urls import reverse

import pytest
from pytest_django.asserts import assertTemplateUsed

from allauth.account.adapter import get_adapter
from allauth.account.authentication import AUTHENTICATION_METHODS_SESSION_KEY


@pytest.fixture
def phone_only_settings(settings_impacting_urls):
    with settings_impacting_urls(
        ACCOUNT_LOGIN_METHODS=("phone",), ACCOUNT_SIGNUP_FIELDS=["phone*"]
    ):
        yield


def test_signup(db, client, phone, sms_outbox, phone_only_settings):
    assert len(sms_outbox) == 0
    resp = client.post(reverse("account_signup"), data={"phone": phone})
    assert resp.status_code == HTTPStatus.FOUND
    assert len(sms_outbox) == 1
    assert resp["location"] == reverse("account_verify_phone")
    resp = client.get(resp["location"])
    assert resp.status_code == HTTPStatus.OK
    resp = client.post(
        reverse("account_verify_phone"), data={"code": sms_outbox[-1]["code"]}
    )
    assert resp.status_code == HTTPStatus.FOUND
    adapter = get_adapter()
    user = adapter.get_user_by_phone(phone)
    phone2, phone_verified = adapter.get_phone(user)
    assert phone_verified
    assert phone == phone2


def test_signup_invalid_attempts(db, client, phone, sms_outbox, phone_only_settings):
    assert len(sms_outbox) == 0
    resp = client.post(reverse("account_signup"), data={"phone": phone})
    assert resp.status_code == HTTPStatus.FOUND
    adapter = get_adapter()
    user = adapter.get_user_by_phone(phone)
    _, phone_verified = adapter.get_phone(user)
    assert not phone_verified
    assert len(sms_outbox) == 1
    assert resp["location"] == reverse("account_verify_phone")
    resp = client.get(resp["location"])
    assert resp.status_code == HTTPStatus.OK
    for i in range(3):
        resp = client.post(reverse("account_verify_phone"), data={"code": "wrong"})
        assert resp.status_code == (HTTPStatus.OK if i < 2 else HTTPStatus.FOUND)


def test_login_sends_code(
    user_with_phone, client, phone_only_settings, phone, sms_outbox
):
    resp = client.post(reverse("account_login"), data={"login": phone})
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_confirm_login_code")
    assert len(sms_outbox) == 1


def test_login_with_verified_phone_and_password(
    client, settings_impacting_urls, phone, user_with_phone, user_password
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*", "password1*"],
        ACCOUNT_LOGIN_METHODS=["phone"],
    ):
        resp = client.post(
            reverse("account_login"), data={"login": phone, "password": user_password}
        )
        assert resp.status_code == HTTPStatus.FOUND
        assert resp["location"] == settings.LOGIN_REDIRECT_URL


def test_login_with_unverified_phone_and_password(
    client, settings_impacting_urls, phone, password_factory, user_factory, sms_outbox
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*", "password1*"],
        ACCOUNT_LOGIN_METHODS=["phone"],
    ):
        password = password_factory()
        user = user_factory(phone=phone, password=password, phone_verified=False)
        resp = client.post(
            reverse("account_login"), data={"login": phone, "password": password}
        )
        assert resp.status_code == HTTPStatus.FOUND
        assert resp["location"] == reverse("account_verify_phone")
        code = sms_outbox[-1]["code"]
        resp = client.post(reverse("account_verify_phone"), data={"code": code})
        assert resp["location"] == settings.LOGIN_REDIRECT_URL
        phone_verified = get_adapter().get_phone(user)
        assert phone_verified == (phone, True)


def test_change_phone(
    auth_client, user, phone_only_settings, phone_factory, sms_outbox
):
    new_phone = phone_factory()
    resp = auth_client.get(reverse("account_change_phone"))
    assert resp.status_code == HTTPStatus.OK
    assertTemplateUsed(resp, "account/phone_change.html")

    resp = auth_client.post(reverse("account_change_phone"), {"phone": new_phone})
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_verify_phone")

    code = sms_outbox[-1]["code"]
    resp = auth_client.get(reverse("account_verify_phone"))
    assert resp.status_code == HTTPStatus.OK
    assertTemplateUsed(resp, "account/confirm_phone_verification_code.html")

    resp = auth_client.post(reverse("account_verify_phone"), {"code": code})
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_change_phone")

    phone_verified = get_adapter().get_phone(user)
    assert phone_verified == (new_phone, True)

    resp = auth_client.get(reverse("account_verify_phone"))
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_change_phone")


def test_change_phone_to_already_existing(
    auth_client, user, phone_only_settings, phone_factory, sms_outbox, user_factory
):
    other_phone = phone_factory()
    user_factory(phone=other_phone)

    resp = auth_client.post(reverse("account_change_phone"), {"phone": other_phone})
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_verify_phone")

    code = sms_outbox[-1]["code"]
    resp = auth_client.get(reverse("account_verify_phone"))
    assert resp.status_code == HTTPStatus.OK
    assertTemplateUsed(resp, "account/confirm_phone_verification_code.html")

    resp = auth_client.post(reverse("account_verify_phone"), {"code": code})
    assert resp.status_code == HTTPStatus.OK
    assert resp.context["form"].errors == {
        "code": ["A user is already registered with this phone number."]
    }


def test_login_by_code_enumeration_prevention(
    db, phone_only_settings, client, phone_factory, sms_outbox
):
    resp = client.post(
        reverse("account_request_login_code"), data={"phone": phone_factory()}
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_confirm_login_code")
    assert "code" not in sms_outbox[-1]
    assert "user_id" not in sms_outbox[-1]


def test_reauthentication(
    settings,
    auth_client,
    user_with_phone,
    phone_factory,
    settings_impacting_urls,
    user_password,
):
    with settings_impacting_urls(
        ACCOUNT_REAUTHENTICATION_REQUIRED=True,
        ACCOUNT_LOGIN_METHODS=("phone",),
        ACCOUNT_SIGNUP_FIELDS=["phone*", "password1*"],
    ):
        new_phone = phone_factory()
        resp = auth_client.post(
            reverse("account_change_phone"),
            {"phone": new_phone},
        )
        assert resp["location"].startswith(reverse("account_reauthenticate"))

        resp = auth_client.get(reverse("account_reauthenticate"))
        assertTemplateUsed(resp, "account/reauthenticate.html")
        resp = auth_client.post(
            reverse("account_reauthenticate"), data={"password": user_password}
        )
        assert resp.status_code == 302

        methods = auth_client.session[AUTHENTICATION_METHODS_SESSION_KEY]
        assert methods[-1] == {"method": "password", "at": ANY, "reauthenticated": True}

        resp = auth_client.post(
            reverse("account_change_phone"),
            {"phone": new_phone},
        )
        assert resp["location"].startswith(reverse("account_verify_phone"))


def test_signup_conflict(db, phone, sms_outbox, phone_only_settings):
    assert len(sms_outbox) == 0
    client = Client()
    resp = client.post(reverse("account_signup"), data={"phone": phone})
    assert resp.status_code == HTTPStatus.FOUND
    assert len(sms_outbox) == 1

    client = Client()
    resp = client.post(reverse("account_signup"), data={"phone": phone})
    assert resp.status_code == HTTPStatus.FOUND
    assert len(sms_outbox) == 2
    assert sms_outbox == [
        {"code": ANY, "phone": phone, "user_id": 1},
        {"phone": phone, "reason": "exists"},
    ]


def test_change_to_already_existing_phone(
    client,
    db,
    settings,
    phone_factory,
    sms_outbox,
    messagesoutbox,
    user_with_phone,
    phone_only_settings,
):
    settings.ACCOUNT_PHONE_VERIFICATION_SUPPORTS_RESEND = True
    settings.ACCOUNT_PHONE_VERIFICATION_SUPPORTS_CHANGE = True
    settings.ACCOUNT_PREVENT_ENUMERATION = True
    new_phone = phone_factory()
    resp = client.post(
        reverse("account_signup"),
        {"phone": new_phone},
    )
    # A user signed up.
    new_user = get_adapter().get_user_by_phone(new_phone)
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_verify_phone")
    assert len(sms_outbox) == 1
    assert sms_outbox[0] == {"code": ANY, "phone": new_phone, "user_id": new_user.pk}
    # assert len(messagesoutbox) == 1
    # assert (
    #     messagesoutbox[-1]["message_template"]
    #     == "account/messages/email_confirmation_sent.txt"
    # )

    # Change to a conflicting phone number
    existing_phone_verified = get_adapter().get_phone(user_with_phone)
    resp = client.post(
        reverse("account_verify_phone"),
        {"action": "change", "phone": existing_phone_verified[0]},
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_verify_phone")
    assert len(sms_outbox) == 2
    assert sms_outbox[1] == {"phone": existing_phone_verified[0], "reason": "exists"}
    # assert len(messagesoutbox) == 2
    # assert (
    #     messagesoutbox[-1]["message_template"]
    #     == "account/messages/email_confirmation_sent.txt"
    # )
    assert get_adapter().get_phone(new_user)[0] == new_phone

    # Change back to new phone
    new_phone2 = phone_factory()
    resp = client.post(
        reverse("account_verify_phone"),
        {"action": "change", "phone": new_phone2},
    )
    assert len(sms_outbox) == 3
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_verify_phone")
    assert sms_outbox[2] == {"code": ANY, "phone": new_phone2, "user_id": ANY}
    # assert len(messagesoutbox) == 3
    # assert (
    #     messagesoutbox[-1]["message_template"]
    #     == "account/messages/email_confirmation_sent.txt"
    # )
    assert get_adapter().get_phone(new_user)[0] == new_phone2

    resp = client.post(
        reverse("account_verify_phone"),
        data={"code": sms_outbox[-1]["code"]},
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert get_adapter().get_phone(new_user) == (new_phone2, True)


def test_signup_invalid_phone(db, client, sms_outbox, phone_only_settings):
    assert len(sms_outbox) == 0
    resp = client.post(reverse("account_signup"), data={"phone": "notaphone"})
    assert resp.status_code == HTTPStatus.OK
    assert resp.context["form"].errors == {
        "phone": ["Enter a phone number including country code (e.g. +1 for the US)."]
    }



from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.urls import reverse

import pytest
from pytest_django.asserts import assertFormError, assertRedirects

from allauth.account import app_settings as account_settings
from allauth.account.models import EmailAddress
from allauth.account.utils import user_email, user_username
from allauth.core import context
from allauth.socialaccount.helpers import complete_social_login
from allauth.socialaccount.models import SocialAccount
from allauth.socialaccount.views import signup


@pytest.fixture
def setup_sociallogin_flow(request_factory):
    def f(client, sociallogin):
        request = request_factory.get("/")
        request.user = AnonymousUser()

        with context.request_context(request):
            resp = complete_social_login(request, sociallogin)
            session = client.session
            for k, v in request.session.items():
                session[k] = v
            session.save()
            return resp

    return f


@pytest.fixture
def email_address_clash(
    request_factory, sociallogin_factory, twitter_provider_settings
):
    def _email_address_clash(username, email):
        User = get_user_model()
        # Some existig user
        exi_user = User()
        user_username(exi_user, "test")
        exi_user_email = "test@example.com"
        user_email(exi_user, exi_user_email)
        exi_user.save()
        EmailAddress.objects.create(
            user=exi_user, email=exi_user_email, verified=True, primary=True
        )

        # A social user being signed up...
        sociallogin = sociallogin_factory(
            provider="twitter", username=username, email=email
        )

        # Signing up, should pop up the social signup form
        request = request_factory.get("/accounts/twitter/login/callback/")
        request.user = AnonymousUser()
        with context.request_context(request):
            resp = complete_social_login(request, sociallogin)
        return request, resp

    return _email_address_clash


def test_email_address_created(
    settings, db, client, setup_sociallogin_flow, sociallogin_factory
):
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True
    settings.ACCOUNT_SIGNUP_FORM_CLASS = None
    settings.ACCOUNT_EMAIL_VERIFICATION = account_settings.EmailVerificationMethod.NONE

    sociallogin = sociallogin_factory(
        email="test@example.com", email_verified=False, username="test"
    )
    setup_sociallogin_flow(client, sociallogin)

    user = get_user_model().objects.get(
        **{account_settings.USER_MODEL_USERNAME_FIELD: "test"}
    )
    assert SocialAccount.objects.filter(user=user, uid=sociallogin.account.uid).exists()
    assert EmailAddress.objects.filter(user=user, email=user_email(user)).exists()


def test_email_address_clash_username_required(
    db, client, settings, email_address_clash
):
    """Test clash on both username and email"""
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = True
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True
    request, resp = email_address_clash("test", "test@example.com")
    assert resp["location"] == reverse("socialaccount_signup")

    # POST different username/email to social signup form
    request.method = "POST"
    request.POST = {"username": "other", "email": "other@example.com"}
    with context.request_context(request):
        resp = signup(request)
    assert resp["location"] == "/accounts/profile/"
    user = get_user_model().objects.get(
        **{account_settings.USER_MODEL_EMAIL_FIELD: "other@example.com"}
    )
    assert user_username(user) == "other"


def test_email_address_clash_username_not_required(db, settings, email_address_clash):
    """Test clash while username is not required"""
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True
    request, resp = email_address_clash("test", "test@example.com")
    assert resp["location"] == reverse("socialaccount_signup")

    # POST email to social signup form (username not present)
    request.method = "POST"
    request.POST = {"email": "other@example.com"}
    with context.request_context(request):
        resp = signup(request)
    assert resp["location"] == "/accounts/profile/"
    user = get_user_model().objects.get(
        **{account_settings.USER_MODEL_EMAIL_FIELD: "other@example.com"}
    )
    assert user_username(user) != "test"


def test_email_address_clash_username_auto_signup(db, settings, email_address_clash):
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True
    # Clash on username, but auto signup still works
    request, resp = email_address_clash("test", "other@example.com")
    assert resp["location"] == "/accounts/profile/"
    user = get_user_model().objects.get(
        **{account_settings.USER_MODEL_EMAIL_FIELD: "other@example.com"}
    )
    assert user_username(user) != "test"


def test_populate_username_in_blacklist(
    db, settings, request_factory, sociallogin_factory, twitter_provider_settings
):
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_USERNAME_BLACKLIST = ["username", "username1", "username2"]
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = True
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True
    request = request_factory.get("/accounts/twitter/login/callback/")
    request.user = AnonymousUser()
    sociallogin = sociallogin_factory(
        provider="twitter", username="username", email="username@example.com"
    )
    with context.request_context(request):
        complete_social_login(request, sociallogin)

    assert request.user.username not in account_settings.USERNAME_BLACKLIST


def test_verified_email_change_at_signup(
    db, client, settings, sociallogin_factory, setup_sociallogin_flow
):
    """
    Test scenario for when the user changes email at social signup. Current
    behavior is that both the unverified and verified email are added, and
    that the user is allowed to pass because he did provide a verified one.
    """
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = False

    sociallogin = sociallogin_factory(email="verified@example.com")
    setup_sociallogin_flow(client, sociallogin)
    resp = client.get(reverse("socialaccount_signup"))
    form = resp.context["form"]
    assert form["email"].value() == "verified@example.com"
    resp = client.post(
        reverse("socialaccount_signup"),
        data={"email": "unverified@example.org"},
    )
    assertRedirects(resp, "/accounts/profile/", fetch_redirect_response=False)
    user = get_user_model().objects.all()[0]
    assert user_email(user) == "verified@example.com"
    assert EmailAddress.objects.filter(
        user=user,
        email="verified@example.com",
        verified=True,
        primary=True,
    ).exists()
    assert EmailAddress.objects.filter(
        user=user,
        email="unverified@example.org",
        verified=False,
        primary=False,
    ).exists()


def test_unverified_email_change_at_signup(
    db, client, settings, sociallogin_factory, setup_sociallogin_flow
):
    """
    Test scenario for when the user changes email at social signup, while
    his provider did not provide a verified email. In that case, email
    verification will kick in. Here, both email addresses are added as
    well.
    """
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = False

    User = get_user_model()
    sociallogin = sociallogin_factory(
        email="unverified@example.com", email_verified=False
    )
    setup_sociallogin_flow(client, sociallogin)
    resp = client.get(reverse("socialaccount_signup"))
    form = resp.context["form"]
    assert form["email"].value() == "unverified@example.com"
    resp = client.post(
        reverse("socialaccount_signup"),
        data={"email": "unverified@example.org"},
    )

    assertRedirects(resp, reverse("account_email_verification_sent"))
    user = User.objects.all()[0]
    assert user_email(user) == "unverified@example.org"
    assert EmailAddress.objects.filter(
        user=user,
        email="unverified@example.com",
        verified=False,
        primary=False,
    ).exists()
    assert EmailAddress.objects.filter(
        user=user,
        email="unverified@example.org",
        verified=False,
        primary=True,
    ).exists()


def test_unique_email_validation_signup(
    db, client, sociallogin_factory, settings, setup_sociallogin_flow
):
    settings.ACCOUNT_PREVENT_ENUMERATION = False
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = False
    User = get_user_model()
    email = "me@example.com"
    user = User.objects.create(email=email)
    EmailAddress.objects.create(email=email, user=user, verified=True)
    sociallogin = sociallogin_factory(email="me@example.com")
    setup_sociallogin_flow(client, sociallogin)
    resp = client.get(reverse("socialaccount_signup"))
    form = resp.context["form"]
    assert form["email"].value() == email
    resp = client.post(reverse("socialaccount_signup"), data={"email": email})
    assertFormError(
        resp.context["form"],
        "email",
        "An account already exists with this email address."
        " Please sign in to that account first, then connect"
        " your Unittest Server account.",
    )


def test_social_account_taken_at_signup(
    db, client, sociallogin_factory, settings, setup_sociallogin_flow
):
    """
    Test scenario for when the user signs up with a social account
    and uses email address in that social account. But upon seeing the
    verification screen, they realize that email address is somehow
    unusable for them, and so backs up and enters a different email
    address (and is forced to choose a new username) while providing
    the same social account token which is owned by their first attempt.
    """
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = True
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.SOCIALACCOUNT_AUTO_SIGNUP = False

    User = get_user_model()
    sociallogin = sociallogin_factory(email="me1@example.com", email_verified=False)
    setup_sociallogin_flow(client, sociallogin)
    resp = client.get(reverse("socialaccount_signup"))
    form = resp.context["form"]
    assert form["email"].value() == "me1@example.com"
    resp = client.post(
        reverse("socialaccount_signup"),
        data={"username": "me1", "email": "me1@example.com"},
    )
    assert resp.status_code == 302
    assert User.objects.count() == 1
    assert SocialAccount.objects.count() == 1

    resp = client.get(reverse("socialaccount_signup"))
    assertRedirects(resp, reverse("account_login"))


def test_email_address_required_missing_from_sociallogin(
    db, settings, sociallogin_factory, client, setup_sociallogin_flow
):
    """Tests that when the email address is missing from the sociallogin email
    verification kicks in.
    """
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True

    sociallogin = sociallogin_factory(with_email=False)
    resp = setup_sociallogin_flow(client, sociallogin)
    assert resp["location"] == reverse("socialaccount_signup")

    resp = client.post(reverse("socialaccount_signup"), {"email": "other@example.org"})
    assert resp["location"] == reverse("account_email_verification_sent")


def test_email_address_conflict_at_social_signup_form(
    db,
    settings,
    user_factory,
    sociallogin_factory,
    client,
    setup_sociallogin_flow,
    mailoutbox,
):
    """Tests that when an already existing email is given at the social signup
    form, enumeration preventation kicks in.
    """
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True

    user = user_factory()
    sociallogin = sociallogin_factory(with_email=False)

    resp = setup_sociallogin_flow(client, sociallogin)
    # Auto signup does not kick in as the `sociallogin` does not have an email.
    assert resp["location"] == reverse("socialaccount_signup")

    # Here, we input the already existing email.
    resp = client.post(reverse("socialaccount_signup"), {"email": user.email})
    assert mailoutbox[0].subject == "[example.com] Account Already Exists"
    assert resp["location"] == reverse("account_email_verification_sent")


def test_email_address_conflict_during_auto_signup(
    db,
    settings,
    user_factory,
    sociallogin_factory,
    client,
    mailoutbox,
    setup_sociallogin_flow,
):
    """Tests that when an already existing email is received from the provider,
    enumeration preventation kicks in.
    """
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True

    user = user_factory()
    sociallogin = sociallogin_factory(email=user.email, with_email=True)
    resp = setup_sociallogin_flow(client, sociallogin)
    assert resp["location"] == reverse("account_email_verification_sent")
    assert mailoutbox[0].subject == "[example.com] Account Already Exists"


def test_email_address_conflict_removes_conflicting_email(
    db,
    settings,
    user_factory,
    sociallogin_factory,
    client,
    mailoutbox,
    setup_sociallogin_flow,
):
    """Tests that when an already existing email is given at the social signup
    form, enumeration preventation kicks in.
    """
    settings.ACCOUNT_EMAIL_REQUIRED = True
    settings.ACCOUNT_UNIQUE_EMAIL = True
    settings.ACCOUNT_USERNAME_REQUIRED = False
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.ACCOUNT_EMAIL_VERIFICATION = "optional"
    settings.SOCIALACCOUNT_AUTO_SIGNUP = True
    settings.SOCIALACCOUNT_EMAIL_AUTHENTICATION = False

    user = user_factory(email_verified=False)
    sociallogin = sociallogin_factory(email=user.email, email_verified=False)

    resp = setup_sociallogin_flow(client, sociallogin)

    # Auto signup does not kick in as the `sociallogin` has a conflicting email.
    assert resp["location"] == reverse("socialaccount_signup")

    # Here, we input the already existing email.
    resp = client.post(reverse("socialaccount_signup"), {"email": "other@email.org"})
    assert mailoutbox[0].subject == "[example.com] Please Confirm Your Email Address"
    assert resp["location"] == settings.LOGIN_REDIRECT_URL
    assert EmailAddress.objects.filter(email=user.email).count() == 1


def test_signup_closed(
    settings,
    db,
    client,
    setup_sociallogin_flow,
    sociallogin_factory,
):
    sociallogin = sociallogin_factory(
        email="test@example.com", email_verified=False, username="test"
    )
    with patch(
        "allauth.socialaccount.adapter.DefaultSocialAccountAdapter.is_open_for_signup"
    ) as iofs:
        iofs.return_value = False
        resp = setup_sociallogin_flow(client, sociallogin)
    assert b"Sign Up Closed" in resp.content
    assert not get_user_model().objects.exists()



from allauth.account.stages import LoginStage
from allauth.core.internal.httpkit import headed_redirect_response, is_headless_request
from allauth.mfa import app_settings
from allauth.mfa.internal.flows import trust
from allauth.mfa.models import Authenticator
from allauth.mfa.utils import is_mfa_enabled
from allauth.mfa.webauthn.internal.flows import did_use_passwordless_login


class AuthenticateStage(LoginStage):
    # NOTE: Duplicated in `allauth.headless.constants.Flow.MFA_AUTHENTICATE`.
    key = "mfa_authenticate"
    urlname = "mfa_authenticate"

    def handle(self):
        response, cont = None, True
        if self._should_handle(self.request):
            self.state["authentication_required"] = True
            response = headed_redirect_response("mfa-authenticate-view")
        return response, cont

    def _should_handle(self, request) -> bool:
        if not is_mfa_enabled(
            self.login.user, request
        ):
            return False
        if did_use_passwordless_login(request):
            return False
        if trust.is_trusted_browser(request, self.login.user):
            return False
        return True


class TrustStage(LoginStage):
    key = "mfa_trust"
    urlname = "mfa_trust"

    def handle(self):
        auth_stage = self.controller.get_stage(AuthenticateStage.key)
        if (
            not app_settings.TRUST_ENABLED
            or not auth_stage
            or not auth_stage.state.get("authentication_required")
        ):
            return None, True
        client = is_headless_request(self.request)
        if client and client == "app":
            # Trust-this-browser relies on cookies.
            return None, True
        response = headed_redirect_response("mfa_trust")
        return response, True



import time

from allauth.socialaccount.internal import statekit


def test_get_oldest_state():
    states = {
        "new": [{"id": "new"}, 300],
        "mid": [{"id": "mid"}, 200],
        "old": [{"id": "old"}, 100],
    }
    state_id, state = statekit.get_oldest_state(states)
    assert state_id == "old"
    assert state["id"] == "old"


def test_get_oldest_state_empty():
    state_id, state = statekit.get_oldest_state({})
    assert state_id is None
    assert state is None


def test_gc_states():
    states = {}
    for i in range(statekit.MAX_STATES + 1):
        states[f"state-{i}"] = [{"i": i}, 1000 + i]
    assert len(states) == statekit.MAX_STATES + 1
    statekit.gc_states(states)
    assert len(states) == statekit.MAX_STATES
    assert "state-0" not in states


def test_stashing(rf):
    request = rf.get("/")
    request.session = {}

    # Stash states with a small delay
    state_id = statekit.stash_state(request, {"foo": "bar"})
    time.sleep(0.001)  # delay for microseconds
    state2_id = statekit.stash_state(request, {"foo2": "bar2"})
    time.sleep(0.001)  # delay for microseconds
    state3_id = statekit.stash_state(request, {"foo3": "bar3"})

    # Unstash last state and check order
    state = statekit.unstash_last_state(request)
    assert state == {"foo3": "bar3"}

    state = statekit.unstash_state(request, state3_id)
    assert state is None
    state = statekit.unstash_state(request, state2_id)
    assert state == {"foo2": "bar2"}
    state = statekit.unstash_state(request, state2_id)
    assert state is None
    state = statekit.unstash_state(request, state_id)
    assert state == {"foo": "bar"}
    state = statekit.unstash_state(request, state_id)
    assert state is None



from http import HTTPStatus

from django.contrib.auth import get_user_model

import pytest

from allauth.account.adapter import get_adapter as get_account_adapter


def test_change_phone_to_same(
    auth_client, user_with_phone, phone, settings_impacting_urls, headless_reverse
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        resp = auth_client.post(
            headless_reverse("headless:account:manage_phone"),
            data={
                "phone": phone,
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.BAD_REQUEST
        assert resp.json() == {
            "errors": [
                {
                    "code": "same_as_current",
                    "message": "The new value must be different from the current one.",
                    "param": "phone",
                }
            ],
            "status": HTTPStatus.BAD_REQUEST,
        }


def test_change_phone(
    auth_client,
    user_with_phone,
    phone,
    settings_impacting_urls,
    headless_reverse,
    phone_factory,
    sms_outbox,
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        new_phone = phone_factory()
        resp = auth_client.post(
            headless_reverse("headless:account:manage_phone"),
            data={
                "phone": new_phone,
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.ACCEPTED
        assert resp.json() == {
            "data": [
                {
                    "phone": new_phone,
                    "verified": False,
                },
            ],
            "status": HTTPStatus.ACCEPTED,
        }

        resp = auth_client.post(
            headless_reverse("headless:account:verify_phone"),
            data={
                "code": sms_outbox[-1]["code"],
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.OK

        resp = auth_client.post(
            headless_reverse("headless:account:verify_phone"),
            data={
                "code": sms_outbox[-1]["code"],
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.CONFLICT

        resp = auth_client.get(
            headless_reverse("headless:account:manage_phone"),
            content_type="application/json",
        )
        assert resp.json() == {
            "data": [
                {
                    "phone": new_phone,
                    "verified": True,
                },
            ],
            "status": 200,
        }


def test_login(
    client,
    user_with_phone,
    phone,
    user_password,
    headless_reverse,
    settings_impacting_urls,
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*", "password1*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        resp = client.post(
            headless_reverse("headless:account:login"),
            data={"phone": phone, "password": user_password},
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.OK


def test_login_by_code(
    client,
    user_with_phone,
    phone,
    headless_reverse,
    settings_impacting_urls,
    sms_outbox,
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        resp = client.post(
            headless_reverse("headless:account:request_login_code"),
            data={
                "phone": phone,
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.UNAUTHORIZED
        code = sms_outbox[-1]["code"]
        resp = client.post(
            headless_reverse("headless:account:confirm_login_code"),
            data={"code": code},
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.OK
        data = resp.json()
        assert data["meta"]["is_authenticated"]


def test_signup(
    db,
    client,
    settings_impacting_urls,
    phone,
    headless_reverse,
    headless_client,
    password_factory,
    sms_outbox,
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*", "password1*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        resp = client.post(
            headless_reverse("headless:account:signup"),
            data={
                "phone": phone,
                "password": password_factory(),
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.UNAUTHORIZED
        pending_flow = [
            flow for flow in resp.json()["data"]["flows"] if flow.get("is_pending")
        ][0]
        assert pending_flow["id"] == "verify_phone"
        assert len(sms_outbox) == 1
        code = sms_outbox[-1]["code"]
        resp = client.post(
            headless_reverse("headless:account:verify_phone"),
            data={
                "code": code,
            },
            content_type="application/json",
        )
        assert resp.json()["status"] == HTTPStatus.OK


def test_reauthentication(
    auth_client,
    user_with_phone,
    phone,
    settings_impacting_urls,
    headless_reverse,
    phone_factory,
    sms_outbox,
):
    with settings_impacting_urls(
        ACCOUNT_REAUTHENTICATION_REQUIRED=True,
        ACCOUNT_SIGNUP_FIELDS=["phone*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        new_phone = phone_factory()
        resp = auth_client.post(
            headless_reverse("headless:account:manage_phone"),
            data={
                "phone": new_phone,
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.UNAUTHORIZED
        assert resp.json()["data"]["flows"] == [{"id": "reauthenticate"}]


def test_change_phone_to_conflicting(
    auth_client,
    user_with_phone,
    phone,
    settings_impacting_urls,
    headless_reverse,
    user_factory,
    phone_factory,
    sms_outbox,
):
    other_phone = phone_factory()
    user_factory(phone=other_phone)
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        resp = auth_client.post(
            headless_reverse("headless:account:manage_phone"),
            data={
                "phone": other_phone,
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.ACCEPTED
        assert resp.json() == {
            "data": [{"phone": other_phone, "verified": False}],
            "status": HTTPStatus.ACCEPTED,
        }
        resp = auth_client.post(
            headless_reverse("headless:account:verify_phone"),
            data={
                "code": sms_outbox[-1]["code"],
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.BAD_REQUEST
        assert resp.json() == {
            "errors": [
                {
                    "code": "phone_taken",
                    "message": "A user is already registered with this phone number.",
                    "param": "code",
                }
            ],
            "status": HTTPStatus.BAD_REQUEST,
        }


def test_change_at_signup(
    db,
    client,
    settings_impacting_urls,
    phone,
    headless_reverse,
    headless_client,
    password_factory,
    sms_outbox,
    phone_factory,
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*", "password1*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
        ACCOUNT_PHONE_VERIFICATION_SUPPORTS_CHANGE=True,
    ):
        resp = client.post(
            headless_reverse("headless:account:signup"),
            data={
                "phone": phone,
                "password": password_factory(),
            },
            content_type="application/json",
        )
        user = get_user_model().objects.last()
        assert resp.status_code == HTTPStatus.UNAUTHORIZED
        assert len(sms_outbox) == 1

        new_phone = phone_factory()
        resp = client.post(
            headless_reverse("headless:account:manage_phone"),
            data={
                "phone": new_phone,
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.ACCEPTED
        assert resp.json() == {
            "data": [
                {
                    "phone": new_phone,
                    "verified": False,
                },
            ],
            "status": HTTPStatus.ACCEPTED,
        }

        assert len(sms_outbox) == 2
        code = sms_outbox[-1]["code"]
        resp = client.post(
            headless_reverse("headless:account:verify_phone"),
            data={
                "code": code,
            },
            content_type="application/json",
        )
        assert resp.json()["status"] == HTTPStatus.OK
        assert get_account_adapter().get_phone(user) == (new_phone, True)


@pytest.mark.parametrize("rate_limit_enabled", [(False,), (True,)])
def test_resend_at_signup(
    db,
    client,
    settings_impacting_urls,
    phone,
    headless_reverse,
    headless_client,
    password_factory,
    sms_outbox,
    phone_factory,
    rate_limit_enabled,
    request,
):
    if rate_limit_enabled:
        request.getfixturevalue("enable_cache")
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*", "password1*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
        ACCOUNT_PHONE_VERIFICATION_SUPPORTS_RESEND=True,
    ):
        resp = client.post(
            headless_reverse("headless:account:signup"),
            data={
                "phone": phone,
                "password": password_factory(),
            },
            content_type="application/json",
        )
        assert resp.status_code == HTTPStatus.UNAUTHORIZED
        assert len(sms_outbox) == 1

        resp = client.post(
            headless_reverse("headless:account:resend_phone_verification_code"),
        )
        if rate_limit_enabled:
            assert resp.status_code == HTTPStatus.TOO_MANY_REQUESTS
        else:
            assert resp.status_code == HTTPStatus.OK
            assert len(sms_outbox) == 2

            old_code = sms_outbox[0]["code"]
            new_code = sms_outbox[1]["code"]
            assert old_code != new_code


def test_manage_phone_anon_no_500(
    client, db, settings_impacting_urls, headless_reverse
):
    with settings_impacting_urls(
        ACCOUNT_SIGNUP_FIELDS=["phone*"],
        ACCOUNT_LOGIN_METHODS=("phone",),
    ):
        resp = client.get(
            headless_reverse("headless:account:manage_phone"),
        )
        assert resp.status_code == HTTPStatus.UNAUTHORIZED



from http import HTTPStatus
from unittest.mock import patch

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import Client
from django.urls import reverse

import pytest

from allauth.account.models import EmailAddress


@pytest.fixture(autouse=True)
def email_verification_settings(settings):
    settings.ACCOUNT_EMAIL_VERIFICATION_BY_CODE_ENABLED = True
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.ACCOUNT_AUTHENTICATION_METHOD = "email"
    return settings


@pytest.mark.parametrize(
    "query,expected_url",
    [
        ("", settings.LOGIN_REDIRECT_URL),
        ("?next=/foo", "/foo"),
    ],
)
def test_signup(
    client,
    db,
    settings,
    password_factory,
    get_last_email_verification_code,
    query,
    expected_url,
    mailoutbox,
):
    password = password_factory()
    resp = client.post(
        reverse("account_signup") + query,
        {
            "username": "johndoe",
            "email": "john@example.com",
            "password1": password,
            "password2": password,
        },
    )
    assert get_user_model().objects.filter(username="johndoe").count() == 1
    code = get_last_email_verification_code(client, mailoutbox)
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    resp = client.get(reverse("account_email_verification_sent"))
    assert resp.status_code == HTTPStatus.OK
    resp = client.post(reverse("account_email_verification_sent"), data={"code": code})
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == expected_url


def test_signup_prevent_enumeration(
    client, db, settings, password_factory, user, mailoutbox
):
    password = password_factory()
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": user.email,
            "password1": password,
            "password2": password,
        },
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    assert not get_user_model().objects.filter(username="johndoe").exists()
    assert mailoutbox[0].subject == "[example.com] Account Already Exists"
    resp = client.get(reverse("account_email_verification_sent"))
    assert resp.status_code == HTTPStatus.OK
    resp = client.post(reverse("account_email_verification_sent"), data={"code": ""})
    assert resp.status_code == HTTPStatus.OK
    assert resp.context["form"].errors == {"code": ["This field is required."]}
    resp = client.post(reverse("account_email_verification_sent"), data={"code": "123"})
    assert resp.status_code == HTTPStatus.OK
    assert resp.context["form"].errors == {"code": ["Incorrect code."]}
    # Max attempts
    resp = client.post(reverse("account_email_verification_sent"), data={"code": "456"})
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_login")


@pytest.mark.parametrize("change_email", (False, True))
def test_add_or_change_email(
    auth_client,
    user,
    get_last_email_verification_code,
    change_email,
    settings,
    mailoutbox,
):
    settings.ACCOUNT_CHANGE_EMAIL = change_email
    settings.ACCOUNT_EMAIL_VERIFICATION_SUPPORTS_RESEND = True
    email = "additional@email.org"
    assert EmailAddress.objects.filter(user=user).count() == 1
    with patch("allauth.account.signals.email_added") as email_added_signal:
        with patch("allauth.account.signals.email_changed") as email_changed_signal:
            resp = auth_client.post(
                reverse("account_email"), {"action_add": "", "email": email}
            )
            assert resp["location"] == reverse("account_email_verification_sent")
            assert not email_added_signal.send.called
            assert not email_changed_signal.send.called
    assert EmailAddress.objects.filter(email=email).count() == 0
    code = get_last_email_verification_code(auth_client, mailoutbox)
    resp = auth_client.get(reverse("account_email_verification_sent"))
    assert resp.status_code == HTTPStatus.OK
    resp = auth_client.post(
        reverse("account_email_verification_sent"), {"action": "resend"}
    )
    assert EmailAddress.objects.filter(email=email).count() == 0
    assert resp.status_code == HTTPStatus.FOUND
    code2 = get_last_email_verification_code(auth_client, mailoutbox)
    assert code != code2
    with patch("allauth.account.signals.email_added") as email_added_signal:
        with patch("allauth.account.signals.email_changed") as email_changed_signal:
            with patch(
                "allauth.account.signals.email_confirmed"
            ) as email_confirmed_signal:
                resp = auth_client.post(
                    reverse("account_email_verification_sent"), data={"code": code2}
                )
                assert resp.status_code == HTTPStatus.FOUND
                assert resp["location"] == settings.LOGIN_REDIRECT_URL
                assert email_added_signal.send.called
                assert email_confirmed_signal.send.called
                assert email_changed_signal.send.called == change_email
    assert EmailAddress.objects.filter(email=email, verified=True).count() == 1
    assert EmailAddress.objects.filter(user=user).count() == (1 if change_email else 2)


def test_email_verification_login_redirect(
    client, db, settings, password_factory, email_verification_settings
):
    password = password_factory()
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": "user@email.org",
            "password1": password,
            "password2": password,
        },
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    resp = client.get(reverse("account_login"))
    assert resp["location"] == reverse("account_email_verification_sent")


def test_email_verification_rate_limits(
    db,
    user_password,
    email_verification_settings,
    settings,
    user_factory,
    password_factory,
    enable_cache,
):
    settings.ACCOUNT_RATE_LIMITS = {"confirm_email": "1/m/key"}
    email = "user@email.org"
    user_factory(email=email, email_verified=False, password=user_password)
    for attempt in range(2):
        resp = Client().post(
            reverse("account_login"),
            {
                "login": email,
                "password": user_password,
            },
        )
        if attempt == 0:
            assert resp.status_code == HTTPStatus.FOUND
            assert resp["location"] == reverse("account_email_verification_sent")
        else:
            assert resp.status_code == HTTPStatus.OK
            assert resp.context["form"].errors == {
                "__all__": ["Too many failed login attempts. Try again later."]
            }


def test_change_email_vs_enumeration_prevention(
    client,
    db,
    settings,
    password_factory,
    get_last_email_verification_code,
    mailoutbox,
    messagesoutbox,
    user,
):
    settings.ACCOUNT_EMAIL_VERIFICATION_SUPPORTS_RESEND = True
    password = password_factory()
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": user.email,
            "password1": password,
            "password2": password,
        },
    )
    # No user signed up.
    assert get_user_model().objects.filter(username="johndoe").count() == 0
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    assert len(mailoutbox) == 1
    assert mailoutbox[0].subject == "[example.com] Account Already Exists"
    assert len(messagesoutbox) == 1
    assert (
        messagesoutbox[-1]["message_template"]
        == "account/messages/email_confirmation_sent.txt"
    )

    # Resend
    resp = client.post(reverse("account_email_verification_sent"), {"action": "resend"})
    assert len(mailoutbox) == 1
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    # No new emails sent.
    assert len(mailoutbox) == 1
    # Yet, pretend we did.
    assert len(messagesoutbox) == 2
    assert (
        messagesoutbox[-1]["message_template"]
        == "account/messages/email_confirmation_sent.txt"
    )


def test_change_to_already_existing_email(
    client,
    db,
    settings,
    password_factory,
    get_last_email_verification_code,
    mailoutbox,
    messagesoutbox,
    user,
):
    settings.ACCOUNT_EMAIL_VERIFICATION_SUPPORTS_RESEND = True
    settings.ACCOUNT_EMAIL_VERIFICATION_SUPPORTS_CHANGE = True
    password = password_factory()
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": "new@user.org",
            "password1": password,
            "password2": password,
        },
    )
    # A user signed up.
    new_user = get_user_model().objects.get(username="johndoe")
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    assert len(mailoutbox) == 1
    assert mailoutbox[0].subject == "[example.com] Please Confirm Your Email Address"
    assert len(messagesoutbox) == 1
    assert (
        messagesoutbox[-1]["message_template"]
        == "account/messages/email_confirmation_sent.txt"
    )

    # Change to a conflicting email
    resp = client.post(
        reverse("account_email_verification_sent"),
        {"action": "change", "email": user.email},
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    assert len(mailoutbox) == 2
    assert mailoutbox[1].subject == "[example.com] Account Already Exists"
    assert len(messagesoutbox) == 2
    assert (
        messagesoutbox[-1]["message_template"]
        == "account/messages/email_confirmation_sent.txt"
    )
    new_user.refresh_from_db()
    assert new_user.email == "new@user.org"

    # Change back to new email
    resp = client.post(
        reverse("account_email_verification_sent"),
        {"action": "change", "email": "new2@user.org"},
    )
    assert len(mailoutbox) == 3
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    assert mailoutbox[2].subject == "[example.com] Please Confirm Your Email Address"
    assert len(messagesoutbox) == 3
    assert (
        messagesoutbox[-1]["message_template"]
        == "account/messages/email_confirmation_sent.txt"
    )
    new_user.refresh_from_db()
    assert new_user.email == "new2@user.org"

    code = get_last_email_verification_code(client, mailoutbox)
    resp = client.post(reverse("account_email_verification_sent"), data={"code": code})
    assert resp.status_code == HTTPStatus.FOUND
    email_address = EmailAddress.objects.filter(user=new_user).get()
    assert email_address.verified
    assert email_address.email == "new2@user.org"


def test_verified_email_decorator__unverified(
    auth_client, mailoutbox, get_last_email_verification_code
):
    EmailAddress.objects.all().update(verified=False)
    resp = auth_client.get(reverse("tests_account_check_verified_email"))
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"].startswith(reverse("account_email_verification_sent"))
    code = get_last_email_verification_code(auth_client, mailoutbox)
    resp = auth_client.post(
        reverse("account_email_verification_sent"),
        data={"code": code, "next": reverse("tests_account_check_verified_email")},
        follow=True,
    )
    assert resp.status_code == HTTPStatus.OK
    assert resp.content == b"VERIFIED"



import json
from typing import Optional
from urllib.parse import parse_qs, quote, urlencode, urlparse, urlunparse

from django import shortcuts
from django.core.exceptions import ImproperlyConfigured
from django.http import (
    HttpRequest,
    HttpResponseRedirect,
    HttpResponseServerError,
    QueryDict,
)
from django.urls import NoReverseMatch, reverse

from allauth import app_settings as allauth_settings


def serialize_request(request):
    return json.dumps(
        {
            "path": request.path,
            "path_info": request.path_info,
            "META": {k: v for k, v in request.META.items() if isinstance(v, str)},
            "GET": request.GET.urlencode(),
            "POST": request.POST.urlencode(),
            "method": request.method,
            "scheme": request.scheme,
        }
    )


def deserialize_request(s, request):
    data = json.loads(s)
    request.GET = QueryDict(data["GET"])
    request.POST = QueryDict(data["POST"])
    request.META = data["META"]
    request.path = data["path"]
    request.path_info = data["path_info"]
    request.method = data["method"]
    request._get_scheme = lambda: data["scheme"]
    return request


def redirect(to):
    try:
        return shortcuts.redirect(to)
    except NoReverseMatch:
        return shortcuts.redirect(f"/{to}")


def del_query_params(url: str, *params: str) -> str:
    parsed_url = urlparse(url)
    query_params = parse_qs(parsed_url.query, keep_blank_values=True)
    for param in params:
        query_params.pop(param, None)
    encoded_query = urlencode(query_params, doseq=True)
    new_url = urlunparse(
        (
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            parsed_url.params,
            encoded_query,
            parsed_url.fragment,
        )
    )
    return new_url


def add_query_params(url: str, params: dict) -> str:
    parsed_url = urlparse(url)
    query_params = parse_qs(parsed_url.query)
    query_params.update(params)
    encoded_query = urlencode(query_params, doseq=True)
    new_url = urlunparse(
        (
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            parsed_url.params,
            encoded_query,
            parsed_url.fragment,
        )
    )
    return new_url


def render_url(request, url_template, **kwargs):
    url = url_template
    for k, v in kwargs.items():
        qi = url.find("?")
        ki = url.find("{" + k + "}")
        if ki < 0:
            raise ImproperlyConfigured(url_template)
        is_query_param = qi >= 0 and ki > qi
        if is_query_param:
            qv = urlencode({"k": v}).partition("k=")[2]
        else:
            qv = quote(v)
        url = url.replace("{" + k + "}", qv)
    p = urlparse(url)
    if not p.netloc:
        url = request.build_absolute_uri(url)
    return url


def default_get_frontend_url(request, urlname, **kwargs):
    from zango.core.utils import get_current_request_url
    # from allauth import app_settings as allauth_settings

    # if allauth_settings.HEADLESS_ENABLED:
    #     from allauth.headless import app_settings as headless_settings

    #     url = headless_settings.FRONTEND_URLS.get(urlname)
    #     if allauth_settings.HEADLESS_ONLY and not url:
    #         raise ImproperlyConfigured(f"settings.HEADLESS_FRONTEND_URLS['{urlname}']")
    return render_url(request, f"{get_current_request_url(request=request)}/{urlname}", **kwargs)
    return None


def get_frontend_url(request, urlname, **kwargs):
    from allauth import app_settings as allauth_settings

    if allauth_settings.HEADLESS_ENABLED:
        from allauth.headless.adapter import get_adapter

        return get_adapter().get_frontend_url(urlname, **kwargs)
    return default_get_frontend_url(request, urlname, **kwargs)


def headed_redirect_response(viewname, query=None):
    """
    In some cases, we're redirecting to a non-headless view. In case of
    headless-only mode, that view clearly does not exist.
    """
    try:
        url = reverse(viewname)
        if query:
            url = add_query_params(url, query)
        return HttpResponseRedirect(url)
    except NoReverseMatch:
        if allauth_settings.HEADLESS_ONLY:
            # The response we would be rendering here is not actually used.
            return HttpResponseServerError()
        raise


def is_headless_request(request: HttpRequest) -> Optional[str]:
    """
    Returns the headless client type (app/browser)in case of a headless
    request.
    """
    return getattr(
        getattr(getattr(request, "allauth", None), "headless", None), "client", None
    )



from allauth.socialaccount.app_settings import QUERY_EMAIL
from allauth.socialaccount.providers.oauth.client import OAuth
from allauth.socialaccount.providers.oauth.views import (
    OAuthAdapter,
    OAuthCallbackView,
    OAuthLoginView,
)


class TwitterAPI(OAuth):
    """
    Verifying twitter credentials
    """

    _base_url = "https://api.twitter.com/1.1/account/verify_credentials.json"
    url = _base_url + "?include_email=true" if QUERY_EMAIL else _base_url

    def get_user_info(self):
        user = self.query(self.url).json()
        return user


class TwitterOAuthAdapter(OAuthAdapter):
    provider_id = "twitter"
    request_token_url = "https://api.twitter.com/oauth/request_token"  # nosec
    access_token_url = "https://api.twitter.com/oauth/access_token"  # nosec
    # Issue #42 -- this one authenticates over and over again...
    # authorize_url = 'https://api.twitter.com/oauth/authorize'
    authorize_url = "https://api.twitter.com/oauth/authenticate"

    def complete_login(self, request, app, token, response):
        client = TwitterAPI(request, app.client_id, app.secret, self.request_token_url)
        extra_data = client.get_user_info()
        return self.get_provider().sociallogin_from_response(request, extra_data)


oauth_login = OAuthLoginView.adapter_view(TwitterOAuthAdapter)
oauth_callback = OAuthCallbackView.adapter_view(TwitterOAuthAdapter)



import pytest

from allauth.mfa.models import Authenticator


@pytest.mark.parametrize("email_verified", [False, True])
def test_get_totp_not_active(auth_client, user, headless_reverse, email_verified):
    resp = auth_client.get(headless_reverse("headless:mfa:manage_totp"))
    if email_verified:
        assert resp.status_code == 404
        data = resp.json()
        assert len(data["meta"]["secret"]) == 32
        assert len(data["meta"]["totp_url"]) == 145
    else:
        assert resp.status_code == 409
        assert resp.json() == {
            "status": 409,
            "errors": [
                {
                    "message": "You cannot activate two-factor authentication until you have verified your email address.",
                    "code": "unverified_email",
                }
            ],
        }


def test_get_totp(
    auth_client,
    user_with_totp,
    headless_reverse,
):
    resp = auth_client.get(headless_reverse("headless:mfa:manage_totp"))
    assert resp.status_code == 200
    data = resp.json()
    assert data["data"]["type"] == "totp"
    assert isinstance(data["data"]["created_at"], float)


def test_deactivate_totp(
    auth_client,
    user_with_totp,
    headless_reverse,
    reauthentication_bypass,
):
    with reauthentication_bypass():
        resp = auth_client.delete(headless_reverse("headless:mfa:manage_totp"))
    assert resp.status_code == 200
    assert not Authenticator.objects.filter(user=user_with_totp).exists()


@pytest.mark.parametrize("email_verified", [False, True])
def test_activate_totp(
    auth_client,
    user,
    headless_reverse,
    reauthentication_bypass,
    settings,
    totp_validation_bypass,
    email_verified,
):
    with reauthentication_bypass():
        with totp_validation_bypass():
            resp = auth_client.post(
                headless_reverse("headless:mfa:manage_totp"),
                data={"code": "42"},
                content_type="application/json",
            )
    if email_verified:
        assert resp.status_code == 200
        assert Authenticator.objects.filter(
            user=user, type=Authenticator.Type.TOTP
        ).exists()
        data = resp.json()
        assert data["data"]["type"] == "totp"
        assert isinstance(data["data"]["created_at"], float)
        assert data["data"]["last_used_at"] is None
    else:
        assert resp.status_code == 400



from http import HTTPStatus

import pytest

from allauth.account.models import (
    EmailAddress,
    EmailConfirmationHMAC,
    get_emailconfirmation_model,
)
from allauth.headless.constants import Flow


def test_verify_email_other_user(auth_client, user, user_factory, headless_reverse):
    other_user = user_factory(email_verified=False)
    email_address = EmailAddress.objects.get(user=other_user, verified=False)
    assert not email_address.verified
    key = EmailConfirmationHMAC(email_address).key
    resp = auth_client.post(
        headless_reverse("headless:account:verify_email"),
        data={"key": key},
        content_type="application/json",
    )
    assert resp.status_code == HTTPStatus.OK
    data = resp.json()
    # We're still authenticated as the user originally logged in, not the
    # other_user.
    assert data["data"]["user"]["id"] == user.pk


@pytest.mark.parametrize(
    "login_on_email_verification,status_code",
    [(False, HTTPStatus.UNAUTHORIZED), (True, HTTPStatus.OK)],
)
def test_auth_unverified_email(
    client,
    user_factory,
    password_factory,
    settings,
    headless_reverse,
    login_on_email_verification,
    status_code,
):
    settings.ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION = login_on_email_verification
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    password = password_factory()
    user = user_factory(email_verified=False, password=password)
    resp = client.post(
        headless_reverse("headless:account:login"),
        data={
            "email": user.email,
            "password": password,
        },
        content_type="application/json",
    )
    assert resp.status_code == HTTPStatus.UNAUTHORIZED
    data = resp.json()
    flows = data["data"]["flows"]
    assert [f for f in flows if f["id"] == Flow.VERIFY_EMAIL][0]["is_pending"]
    emailaddress = EmailAddress.objects.filter(user=user, verified=False).get()
    key = get_emailconfirmation_model().create(emailaddress).key
    resp = client.post(
        headless_reverse("headless:account:verify_email"),
        data={"key": key},
        content_type="application/json",
    )
    assert resp.status_code == status_code


def test_verify_email_bad_key(
    client, settings, password_factory, user_factory, headless_reverse
):
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    password = password_factory()
    user = user_factory(email_verified=False, password=password)
    resp = client.post(
        headless_reverse("headless:account:login"),
        data={
            "email": user.email,
            "password": password,
        },
        content_type="application/json",
    )
    assert resp.status_code == HTTPStatus.UNAUTHORIZED
    resp = client.get(
        headless_reverse("headless:account:verify_email"),
        data={"key": "bad"},
        content_type="application/json",
    )
    assert resp.status_code == HTTPStatus.BAD_REQUEST
    resp = client.post(
        headless_reverse("headless:account:verify_email"),
        data={"key": "bad"},
        content_type="application/json",
    )
    assert resp.status_code == HTTPStatus.BAD_REQUEST
    assert resp.json() == {
        "status": HTTPStatus.BAD_REQUEST,
        "errors": [
            {
                "code": "invalid_or_expired_key",
                "param": "key",
                "message": "Invalid or expired key.",
            }
        ],
    }



import json

from django.conf import settings
from django.urls import reverse

import pytest
from pytest_django.asserts import assertRedirects

from allauth.account import app_settings


@pytest.mark.parametrize(
    "headers,ajax_expected",
    [
        ({}, False),
        ({"HTTP_X_REQUESTED_WITH": "XMLHttpRequest"}, True),
        ({"HTTP_ACCEPT": "application/json"}, True),
    ],
)
def test_ajax_headers(db, client, headers, ajax_expected):
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": "john@example.org",
            "email2": "john@example.org",
            "password1": "johndoe",
            "password2": "johndoe",
        },
        **headers,
    )
    if ajax_expected:
        assert resp.status_code == 200
        assert resp.json()["location"] == settings.LOGIN_REDIRECT_URL
        assert resp.json()["location"] == settings.LOGIN_REDIRECT_URL
    else:
        assert resp.status_code == 302
        assertRedirects(
            resp, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False
        )


def test_ajax_password_reset(client, user, mailoutbox):
    resp = client.post(
        reverse("account_reset_password"),
        data={"email": user.email},
        HTTP_X_REQUESTED_WITH="XMLHttpRequest",
    )
    assert len(mailoutbox) == 1
    assert mailoutbox[0].to == [user.email]
    assert resp["content-type"] == "application/json"


def test_ajax_login_fail(client, db):
    resp = client.post(
        reverse("account_login"),
        {},
        HTTP_X_REQUESTED_WITH="XMLHttpRequest",
    )
    assert resp.status_code == 400
    json.loads(resp.content.decode("utf8"))
    # TODO: Actually test something


def test_ajax_login_success(settings, user, user_password, client):
    settings.ACCOUNT_EMAIL_VERIFICATION = app_settings.EmailVerificationMethod.OPTIONAL
    resp = client.post(
        reverse("account_login"),
        {"login": user.username, "password": user_password},
        HTTP_X_REQUESTED_WITH="XMLHttpRequest",
    )
    assert resp.status_code == 200
    data = json.loads(resp.content.decode("utf8"))
    assert data["location"] == "/accounts/profile/"



from django.test import TestCase
from django.test.utils import override_settings

from allauth.socialaccount.providers.nextcloud.provider import NextCloudProvider
from tests.apps.socialaccount.base import OAuth2TestsMixin
from tests.mocking import MockedResponse


@override_settings(
    SOCIALACCOUNT_PROVIDERS={"nextcloud": {"SERVER": "https://nextcloud.example.org"}}
)
class NextCloudTests(OAuth2TestsMixin, TestCase):
    provider_id = NextCloudProvider.id

    def get_login_response_json(self, with_refresh_token=True):
        return (
            super(NextCloudTests, self)
            .get_login_response_json(with_refresh_token=with_refresh_token)
            .replace("uid", "user_id")
        )

    def get_mocked_response(self):
        return MockedResponse(
            200,
            """
{
  "ocs": {
    "meta": {
      "status": "ok",
      "statuscode": 100,
      "message": "OK",
      "totalitems": "",
      "itemsperpage": ""
    },
    "data": {
      "enabled": true,
      "storageLocation": "\\/var\\/www\\/html\\/data\\/pennersr",
      "id": "pennersr",
      "lastLogin": 1730973409000,
      "backend": "Database",
      "subadmin": [],
      "quota": {
        "free": 9159623057408,
        "used": 1585107741,
        "total": 9161208165149,
        "relative": 0.02,
        "quota": -3
      },
      "email": "batman@wayne.com",
      "displayname": "pennersr",
      "phone": "",
      "address": "",
      "website": "",
      "twitter": "",
      "groups": [
        "admin"
      ],
      "language": "nl",
      "locale": ""
    }
  }
}
""",
        )

    def get_expected_to_str(self):
        return "batman@wayne.com"



import json
from typing import Dict, Optional, Type, Union

from django.http import HttpResponseBadRequest
from django.views.generic import View

from allauth.core.exceptions import ImmediateHttpResponse
from allauth.headless.internal.restkit.inputs import Input
from allauth.headless.internal.restkit.response import ErrorResponse


class RESTView(View):
    input_class: Union[Optional[Dict[str, Type[Input]]], Type[Input]] = None
    handle_json_input = True

    def dispatch(self, request, *args, **kwargs):
        return self.handle(request, *args, **kwargs)

    def handle(self, request, *args, **kwargs):
        if self.handle_json_input and request.method != "GET":
            self.data = self._parse_json(request)
            response = self.handle_input(self.data)
            if response:
                return response
        return super().dispatch(request, *args, **kwargs)

    def get_input_class(self):
        input_class = self.input_class
        if isinstance(input_class, dict):
            input_class = input_class.get(self.request.method)
        return input_class

    def get_input_kwargs(self):
        return {}

    def handle_input(self, data):
        input_class = self.get_input_class()
        if not input_class:
            return
        input_kwargs = self.get_input_kwargs()
        if data is None:
            # Make form bound on empty POST
            data = {}
        self.input = input_class(data=data, **input_kwargs)
        if not self.input.is_valid():
            return self.handle_invalid_input(self.input)

    def handle_invalid_input(self, input):
        return ErrorResponse(self.request, input=input)

    def _parse_json(self, request):
        if request.method == "GET" or not request.body:
            return
        try:
            return json.loads(request.body.decode("utf8"))
        except (UnicodeDecodeError, json.JSONDecodeError):
            raise ImmediateHttpResponse(response=HttpResponseBadRequest())



from typing import Optional

from allauth.socialaccount import app_settings
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.providers.oauth2.views import (
    OAuth2Adapter,
    OAuth2CallbackView,
    OAuth2LoginView,
)


class GitHubOAuth2Adapter(OAuth2Adapter):
    provider_id = "github"
    settings = app_settings.PROVIDERS.get(provider_id, {})

    if "GITHUB_URL" in settings:
        web_url = settings.get("GITHUB_URL").rstrip("/")
        api_url = "{0}/api/v3".format(web_url)
    else:
        web_url = "https://github.com"
        api_url = "https://api.github.com"

    access_token_url = "{0}/login/oauth/access_token".format(web_url)
    authorize_url = "{0}/login/oauth/authorize".format(web_url)
    profile_url = "{0}/user".format(api_url)
    emails_url = "{0}/user/emails".format(api_url)

    def complete_login(self, request, app, token, **kwargs):
        headers = {"Authorization": "token {}".format(token.token)}
        resp = (
            get_adapter().get_requests_session().get(self.profile_url, headers=headers)
        )
        resp.raise_for_status()
        extra_data = resp.json()
        if app_settings.QUERY_EMAIL:
            if emails := self.get_emails(headers):
                extra_data["emails"] = emails
        return self.get_provider().sociallogin_from_response(request, extra_data)

    def get_emails(self, headers) -> Optional[list]:
        resp = (
            get_adapter().get_requests_session().get(self.emails_url, headers=headers)
        )
        # https://api.github.com/user/emails -- 404 is documented to occur.
        if resp.status_code == 404:
            return None
        resp.raise_for_status()
        return resp.json()


oauth2_login = OAuth2LoginView.adapter_view(GitHubOAuth2Adapter)
oauth2_callback = OAuth2CallbackView.adapter_view(GitHubOAuth2Adapter)



from datetime import timedelta
from http import HTTPStatus
from unittest.mock import Mock

from django.contrib.auth import SESSION_KEY, get_user_model
from django.core.cache import cache
from django.urls import reverse
from django.utils.timezone import now

import pytest
from pytest_django.asserts import (
    assertRedirects,
    assertTemplateNotUsed,
    assertTemplateUsed,
)

from allauth.account import app_settings
from allauth.account.adapter import DefaultAccountAdapter
from allauth.account.models import (
    EmailAddress,
    EmailConfirmation,
    EmailConfirmationHMAC,
)
from allauth.account.signals import user_logged_in


class TestEmailVerificationAdapter(DefaultAccountAdapter):
    SIGNUP_REDIRECT_URL = "/foobar"

    def get_signup_redirect_url(self, request):
        return self.SIGNUP_REDIRECT_URL


@pytest.mark.parametrize(
    "adapter,query,expected_location",
    [
        (None, "", app_settings.SIGNUP_REDIRECT_URL),
        (None, "?next=/foo", "/foo"),
        (
            "tests.apps.account.test_email_verification.TestEmailVerificationAdapter",
            "",
            TestEmailVerificationAdapter.SIGNUP_REDIRECT_URL,
        ),
    ],
)
def test_login_on_verification(
    adapter, client, db, query, expected_location, password_factory, settings
):
    settings.ACCOUNT_EMAIL_VERIFICATION = app_settings.EmailVerificationMethod.MANDATORY
    settings.ACCOUNT_EMAIL_CONFIRMATION_HMAC = True
    settings.ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION = True
    if adapter:
        settings.ACCOUNT_ADAPTER = adapter
    password = password_factory()
    resp = client.post(
        reverse("account_signup"),
        data={
            "username": "john",
            "email": "a@a.com",
            "password1": password,
            "password2": password,
        },
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")

    resp = client.get(resp["location"])
    assert resp.status_code == HTTPStatus.OK

    email = EmailAddress.objects.get(email="a@a.com")
    key = EmailConfirmationHMAC(email).key

    receiver_mock = Mock()  # we've logged if signal was called
    user_logged_in.connect(receiver_mock)

    resp = client.post(reverse("account_confirm_email", args=[key]) + query)
    assert resp["location"] == expected_location
    email = EmailAddress.objects.get(pk=email.pk)
    assert email.verified

    receiver_mock.assert_called_once_with(
        sender=get_user_model(),
        request=resp.wsgi_request,
        response=resp,
        user=email.user,
        signal=user_logged_in,
    )

    user_logged_in.disconnect(receiver_mock)


def test_email_verification_failed(settings, user_factory, client):
    settings.ACCOUNT_EMAIL_CONFIRMATION_HMAC = False
    user_factory(email_verified=True, email="foo@bar.org")
    unverified_user = user_factory(email_verified=False, email="foo@bar.org")
    email_address = EmailAddress.objects.get_for_user(
        unverified_user, unverified_user.email
    )
    assert not email_address.verified
    confirmation = EmailConfirmation.objects.create(
        email_address=email_address,
        key="dummy",
        sent=now(),
    )
    resp = client.post(reverse("account_confirm_email", args=[confirmation.key]))
    assertTemplateUsed(resp, "account/messages/email_confirmation_failed.txt")


def test_email_verification_mandatory(settings, db, client, mailoutbox, enable_cache):
    settings.ACCOUNT_EMAIL_CONFIRMATION_HMAC = False
    settings.ACCOUNT_EMAIL_CONFIRMATION_COOLDOWN = 10
    settings.ACCOUNT_EMAIL_VERIFICATION = app_settings.EmailVerificationMethod.MANDATORY
    # Signup
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": "john@example.com",
            "password1": "johndoe",
            "password2": "johndoe",
        },
        follow=True,
    )
    assert resp.status_code == HTTPStatus.OK
    assert mailoutbox[0].to == ["john@example.com"]
    assert mailoutbox[0].body.find("http://") > 0
    assert len(mailoutbox) == 1
    assertTemplateUsed(
        resp,
        "account/verification_sent.%s" % app_settings.TEMPLATE_EXTENSION,
    )
    # Attempt to login, unverified
    for attempt in [1, 2]:
        resp = client.post(
            reverse("account_login"),
            {"login": "johndoe", "password": "johndoe"},
            follow=True,
        )
        # is_active is controlled by the admin to manually disable
        # users. I don't want this flag to flip automatically whenever
        # users verify their email addresses.
        assert (
            get_user_model().objects.filter(username="johndoe", is_active=True).exists()
        )

        assertTemplateUsed(
            resp,
            "account/verification_sent." + app_settings.TEMPLATE_EXTENSION,
        )
        # Attempt 1: no mail is sent due to cool-down ,
        # but there was already a mail in the outbox.
        assert len(mailoutbox) == attempt
        assert (
            EmailConfirmation.objects.filter(
                email_address__email="john@example.com"
            ).count()
            == attempt
        )
        # Wait for cooldown -- wipe cache (incl. rate limits)
        cache.clear()
        # if we don't wipe the session, login will redirect to pending stage...
        client.logout()

    # Verify, and re-attempt to login.
    confirmation = EmailConfirmation.objects.filter(
        email_address__user__username="johndoe"
    )[:1].get()
    resp = client.get(reverse("account_confirm_email", args=[confirmation.key]))
    assertTemplateUsed(
        resp, "account/email_confirm.%s" % app_settings.TEMPLATE_EXTENSION
    )
    client.post(reverse("account_confirm_email", args=[confirmation.key]))
    resp = client.post(
        reverse("account_login"),
        {"login": "johndoe", "password": "johndoe"},
    )
    assertRedirects(resp, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)


def test_optional_email_verification(settings, client, db, mailoutbox):
    settings.ACCOUNT_SIGNUP_REDIRECT_URL = "/accounts/welcome/"
    settings.ACCOUNT_EMAIL_VERIFICATION = app_settings.EmailVerificationMethod.OPTIONAL
    settings.ACCOUNT_SIGNUP_PASSWORD_ENTER_TWICE = False
    # Signup
    client.get(reverse("account_signup"))
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": "john@example.com",
            "password1": "johndoe",
        },
    )
    # Logged in
    assertRedirects(
        resp, settings.ACCOUNT_SIGNUP_REDIRECT_URL, fetch_redirect_response=False
    )
    assert mailoutbox[0].to == ["john@example.com"]
    assert len(mailoutbox) == 1
    # Logout & login again
    client.logout()
    # Wait for cooldown
    EmailConfirmation.objects.update(sent=now() - timedelta(days=1))
    # Signup
    resp = client.post(
        reverse("account_login"),
        {"login": "johndoe", "password": "johndoe"},
    )
    assertRedirects(resp, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)
    assert mailoutbox[0].to == ["john@example.com"]
    # There was an issue that we sent out email confirmation mails
    # on each login in case of optional verification. Make sure
    # this is not the case:
    assert len(mailoutbox) == 1


def test_email_verification_hmac(settings, client, user_factory, mailoutbox, rf):
    settings.ACCOUNT_EMAIL_CONFIRMATION_HMAC = True
    user = user_factory(email_verified=False)
    email = EmailAddress.objects.get_for_user(user, user.email)
    confirmation = EmailConfirmationHMAC(email)
    request = rf.get("/")
    confirmation.send(request=request)
    assert len(mailoutbox) == 1
    client.post(reverse("account_confirm_email", args=[confirmation.key]))
    email = EmailAddress.objects.get(pk=email.pk)
    assert email.verified


def test_email_verification_hmac_timeout(
    settings, user_factory, client, mailoutbox, rf
):
    settings.ACCOUNT_EMAIL_CONFIRMATION_HMAC = True
    settings.ACCOUNT_EMAIL_CONFIRMATION_EXPIRE_DAYS = 0
    user = user_factory(email_verified=False)
    email = EmailAddress.objects.get_for_user(user, user.email)
    confirmation = EmailConfirmationHMAC(email)
    request = rf.get("/")
    confirmation.send(request=request)
    assert len(mailoutbox) == 1
    client.post(reverse("account_confirm_email", args=[confirmation.key]))
    email = EmailAddress.objects.get(pk=email.pk)
    assert not email.verified


def test_verify_email_with_another_user_logged_in(
    settings, user_factory, client, mailoutbox
):
    """Test the email verification view. If User B clicks on an email
    verification link while logged in as User A, ensure User A gets
    logged out."""
    settings.ACCOUNT_LOGIN_METHODS = {app_settings.AuthenticationMethod.EMAIL}
    user = user_factory(email_verified=False)
    client.force_login(user)
    client.post(reverse("account_email"), {"email": user.email, "action_send": ""})
    assert len(mailoutbox) == 1
    assert mailoutbox[0].to == [user.email]
    client.logout()
    body = mailoutbox[0].body
    assert body.find("http://") > 0

    user2 = user_factory(email_verified=False, password="doe")
    resp = client.post(
        reverse("account_login"),
        {
            "login": user2.email,
            "password": "doe",
        },
    )
    assert user2 == resp.context["user"]

    url = body[body.find("/accounts/confirm-email/") :].split()[0]
    resp = client.post(url)

    assertTemplateUsed(resp, "account/messages/logged_out.txt")
    assertTemplateUsed(resp, "account/messages/email_confirmed.txt")

    assertRedirects(resp, settings.LOGIN_URL, fetch_redirect_response=False)


def test_verify_email_with_same_user_logged_in(
    settings, user_factory, client, mailoutbox
):
    """If the user clicks on an email verification link while logged in, ensure
    the user stays logged in.
    """
    settings.ACCOUNT_LOGIN_METHODS = {app_settings.AuthenticationMethod.EMAIL}
    user = user_factory(email_verified=False)
    client.force_login(user)
    client.post(reverse("account_email"), {"email": user.email, "action_send": ""})
    assert len(mailoutbox) == 1
    assert mailoutbox[0].to == [user.email]

    body = mailoutbox[0].body
    assert body.find("http://") > 0

    url = body[body.find("/accounts/confirm-email/") :].split()[0]
    resp = client.post(url)

    assertTemplateNotUsed(resp, "account/messages/logged_out.txt")
    assertTemplateUsed(resp, "account/messages/email_confirmed.txt")

    assertRedirects(resp, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)

    assert user == resp.wsgi_request.user


def test_verify_logs_out_user(auth_client, settings, user, user_factory):
    """
    When a user is signed in, and you follow an email confirmation link of
    another user within the same browser/session, be sure to sign out the signed
    in user.
    """
    settings.ACCOUNT_CONFIRM_EMAIL_ON_GET = False
    confirming_user = user_factory(email_verified=False)
    assert auth_client.session[SESSION_KEY] == str(user.pk)
    email = EmailAddress.objects.get(user=confirming_user, verified=False)
    auth_client.get(
        reverse(
            "account_confirm_email", kwargs={"key": EmailConfirmationHMAC(email).key}
        )
    )
    assert not auth_client.session.get(SESSION_KEY)


def test_email_verification_login_redirect(client, db, settings, password_factory):
    settings.ACCOUNT_EMAIL_VERIFICATION = app_settings.EmailVerificationMethod.MANDATORY
    password = password_factory()
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": "user@email.org",
            "password1": password,
            "password2": password,
        },
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    resp = client.get(reverse("account_login"))
    assert resp.status_code == HTTPStatus.OK


def test_email_verification_redirect_url(client, db, settings, user_password):
    settings.ACCOUNT_EMAIL_VERIFICATION = app_settings.EmailVerificationMethod.MANDATORY
    settings.ACCOUNT_EMAIL_CONFIRMATION_ANONYMOUS_REDIRECT_URL = "/foobar"
    settings.ACCOUNT_CONFIRM_EMAIL_ON_GET = True
    email = "user@email.org"
    resp = client.post(
        reverse("account_signup"),
        {
            "username": "johndoe",
            "email": email,
            "password1": user_password,
            "password2": user_password,
        },
    )
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == reverse("account_email_verification_sent")
    confirmation = EmailConfirmationHMAC(EmailAddress.objects.get(email=email))
    resp = client.get(reverse("account_confirm_email", args=[confirmation.key]))
    assert resp.status_code == HTTPStatus.FOUND
    assert resp["location"] == "/foobar"


def test_verified_email_decorator__unverified(auth_client):
    EmailAddress.objects.all().update(verified=False)
    resp = auth_client.get(reverse("tests_account_check_verified_email"))
    assert resp.status_code == HTTPStatus.OK
    assertTemplateUsed(resp, "account/verified_email_required.html")


def test_verified_email_decorator__verified(auth_client):
    EmailAddress.objects.all().update(verified=True)
    resp = auth_client.get(reverse("tests_account_check_verified_email"))
    assert resp.status_code == HTTPStatus.OK
    assert resp.content == b"VERIFIED"



from django.urls import reverse


def test_openapi_json(client):
    resp = client.get(reverse("headless:openapi_json"))
    assert resp.status_code == 200
    data = resp.json()
    assert data["openapi"] == "3.0.3"
    assert data["info"]["description"].startswith("# Introduction")



from django.contrib import admin
from django.urls import include, path
from django.views.generic.base import TemplateView

from allauth.account.decorators import secure_admin_login


admin.autodiscover()
admin.site.login = secure_admin_login(admin.site.login)

urlpatterns = [
    path("", TemplateView.as_view(template_name="index.html")),
    path("accounts/", include("allauth.urls")),
    path("accounts/profile/", TemplateView.as_view(template_name="profile.html")),
    path("admin/", admin.site.urls),
    path("i18n/", include("django.conf.urls.i18n")),
    path("", include("allauth.idp.urls")),
]



from unittest.mock import patch

from django.urls import reverse

import pytest
from pytest_django.asserts import assertTemplateUsed

from allauth.socialaccount.internal import flows
from allauth.socialaccount.models import SocialAccount
from allauth.socialaccount.providers.base.constants import AuthProcess


@pytest.mark.parametrize("reauthentication_required", [False, True])
def test_disconnect(auth_client, user, settings, mailoutbox, reauthentication_required):
    settings.ACCOUNT_EMAIL_NOTIFICATIONS = True
    settings.ACCOUNT_REAUTHENTICATION_REQUIRED = reauthentication_required
    account_to_del = SocialAccount.objects.create(
        uid="123", provider="other-server", user=user
    )
    account_to_keep = SocialAccount.objects.create(
        uid="456", provider="other-server", user=user
    )
    resp = auth_client.get(reverse("socialaccount_connections"))
    assertTemplateUsed(resp, "socialaccount/connections.html")
    resp = auth_client.post(
        reverse("socialaccount_connections"), {"account": account_to_del.pk}
    )
    if reauthentication_required:
        assert SocialAccount.objects.filter(pk=account_to_del.pk).exists()
        assert SocialAccount.objects.filter(pk=account_to_keep.pk).exists()
    else:
        assert not SocialAccount.objects.filter(pk=account_to_del.pk).exists()
        assert SocialAccount.objects.filter(pk=account_to_keep.pk).exists()
        assert len(mailoutbox) == 1
        assert mailoutbox[0].subject == "[example.com] Third-Party Account Disconnected"


def test_connect_with_reauthentication(
    auth_client, user, provider_callback_response, settings, user_password
):
    settings.ACCOUNT_REAUTHENTICATION_REQUIRED = True
    resp = provider_callback_response(auth_client, process="connect")
    assert not SocialAccount.objects.filter(user=user).exists()
    assert resp.status_code == 302
    assert resp["location"] == reverse("account_reauthenticate")
    resp = auth_client.post(resp["location"], {"password": user_password})
    assert resp.status_code == 302
    assert resp["location"] == reverse("socialaccount_connections")
    assert SocialAccount.objects.filter(user=user).exists()


def test_connect(
    auth_client, user, provider_callback_response, settings, user_password, mailoutbox
):
    settings.ACCOUNT_EMAIL_NOTIFICATIONS = True
    settings.ACCOUNT_REAUTHENTICATION_REQUIRED = False
    resp = provider_callback_response(auth_client, process="connect")
    assert resp.status_code == 302
    assert SocialAccount.objects.filter(user=user).exists()
    assert resp["location"] == reverse("socialaccount_connections")
    assert len(mailoutbox) == 1
    assert mailoutbox[0].subject == "[example.com] Third-Party Account Connected"


@pytest.mark.parametrize(
    "email_authentication,account_exists, expected_action",
    [
        (False, False, "added"),
        (False, True, "updated"),
        (True, False, "added"),
        (True, True, "updated"),
    ],
)
def test_connect_vs_email_authentication(
    request_factory,
    sociallogin_factory,
    user,
    settings,
    email_authentication,
    account_exists,
    expected_action,
):
    settings.SOCIALACCOUNT_EMAIL_AUTHENTICATION = email_authentication
    sociallogin = sociallogin_factory(email=user.email, provider="unittest-server")
    if account_exists:
        account = sociallogin.account
        account.user = user
        account.save()

    sociallogin.state["process"] = AuthProcess.CONNECT
    request = request_factory.get("/")
    request.user = user
    with patch(
        "allauth.account.adapter.DefaultAccountAdapter.add_message"
    ) as add_message:
        flows.login.complete_login(request, sociallogin)
        assert add_message.call_args[1]["message_context"]["action"] == expected_action
    assert SocialAccount.objects.filter(user=user, uid=sociallogin.account.uid).exists()



import datetime
from typing import Any, Dict, Tuple

from django import forms

from allauth.account.fields import EmailField


FIELD_MAPPING = {
    forms.CharField: {"type": "string"},
    forms.IntegerField: {"type": "integer"},
    forms.FloatField: {"type": "number"},
    forms.BooleanField: {"type": "boolean"},
    forms.DateField: {"type": "string", "format": "date"},
    forms.DateTimeField: {"type": "string", "format": "date-time"},
    forms.EmailField: {"type": "string", "format": "email"},
    EmailField: {"type": "string", "format": "email"},
    forms.URLField: {"type": "string", "format": "uri"},
    forms.DecimalField: {
        "type": "string",
        "format": "decimal",
        "pattern": r"^\d+(\.\d+)?$",
    },
}


def spec_for_field(field: forms.Field) -> Dict[str, Any]:
    field_spec: Dict[str, Any] = FIELD_MAPPING.get(type(field), {"type": "string"})
    field_spec = dict(field_spec)
    if hasattr(field, "max_length") and field.max_length:
        field_spec["maxLength"] = field.max_length
    if hasattr(field, "min_length") and field.min_length:
        field_spec["minLength"] = field.min_length
    if hasattr(field, "help_text") and field.help_text:
        field_spec["description"] = field.help_text
    return field_spec


def spec_for_dataclass(dc) -> Tuple[dict, dict]:
    example = {}
    props = {}
    for field_id, field in dc.__dataclass_fields__.items():
        example[field_id] = field.metadata["example"]
        descriptor = {
            "description": field.metadata["description"],
            "example": field.metadata["example"],
        }
        if field.type is str:
            descriptor.update({"type": "string"})
        elif field.type is int:
            descriptor.update({"type": "integer"})
        elif field.type is float:
            descriptor.update({"type": "number", "format": "float"})
        elif field.type is bool:
            descriptor.update({"type": "boolean"})
        elif field.type is datetime.datetime:
            descriptor.update({"type": "string", "format": "date-time"})
        elif field.type is datetime.date:
            descriptor.update({"type": "string", "format": "date"})
        elif field.type is list:
            descriptor.update({"type": "array"})
        elif field.type is dict:
            descriptor.update({"type": "object"})
        else:
            descriptor.update({"type": "string"})
        props[field_id] = descriptor
    schema = {"type": "object", "properties": props}
    return schema, example



from django.test import TestCase

from allauth.socialaccount.providers.amazon.provider import AmazonProvider
from tests.apps.socialaccount.base import OAuth2TestsMixin
from tests.mocking import MockedResponse


class AmazonTests(OAuth2TestsMixin, TestCase):
    provider_id = AmazonProvider.id

    def get_mocked_response(self):
        return MockedResponse(
            200,
            """
        {
          "Profile":{
                        "CustomerId":"amzn1.account.K2LI23KL2LK2",
                        "Name":"John Doe",
                        "PrimaryEmail":"johndoe@example.com"
                    }
        }""",
        )

    def get_expected_to_str(self):
        return "johndoe@example.com"



from django.urls import reverse

import pytest
from pytest_django.asserts import assertTemplateUsed

from allauth.socialaccount.adapter import get_adapter


@pytest.mark.parametrize(
    "samesite_strict,did_already_redirect,expect_redirect",
    [
        (True, False, True),
        (True, True, False),
        (False, False, False),
    ],
)
def test_samesite_strict(
    client,
    samesite_strict,
    settings,
    google_provider_settings,
    did_already_redirect,
    expect_redirect,
    db,
):
    settings.SESSION_COOKIE_SAMESITE = "Strict" if samesite_strict else "Lax"
    query = "?state=123"
    resp = client.get(
        reverse("google_callback") + query + ("&_redir" if did_already_redirect else "")
    )
    if expect_redirect:
        assertTemplateUsed(resp, "socialaccount/login_redirect.html")
        assert (
            resp.context["redirect_to"]
            == reverse("google_callback") + query + "&_redir="
        )
    else:
        assertTemplateUsed(resp, "socialaccount/authentication_error.html")


@pytest.mark.parametrize("pkce_enabled", [False, True])
def test_config_from_app_settings(
    google_provider_settings, rf, db, settings, pkce_enabled
):
    settings.SOCIALACCOUNT_PROVIDERS["google"]["APPS"][0]["settings"] = {
        "scope": ["this", "that"],
        "auth_params": {"x": "y"},
        "oauth_pkce_enabled": pkce_enabled,
    }
    settings.SOCIALACCOUNT_PROVIDERS["google"]["SCOPE"] = ["not-this"]
    settings.SOCIALACCOUNT_PROVIDERS["google"]["AUTH_PARAMS"] = {"not": "this"}
    provider = get_adapter().get_provider(rf.get("/"), "google")
    assert provider.get_scope() == ["this", "that"]
    assert provider.get_auth_params() == {"x": "y"}
    assert ("code_verifier" in provider.get_pkce_params().keys()) == pkce_enabled


@pytest.mark.parametrize("pkce_enabled", [False, True])
def test_config_from_provider_config(
    google_provider_settings, rf, db, settings, pkce_enabled
):
    settings.SOCIALACCOUNT_PROVIDERS["google"]["SCOPE"] = ["some-scope"]
    settings.SOCIALACCOUNT_PROVIDERS["google"]["AUTH_PARAMS"] = {"auth": "param"}
    settings.SOCIALACCOUNT_PROVIDERS["google"]["OAUTH_PKCE_ENABLED"] = pkce_enabled
    provider = get_adapter().get_provider(rf.get("/"), "google")
    assert provider.get_scope() == ["some-scope"]
    assert provider.get_auth_params() == {"auth": "param"}
    assert ("code_verifier" in provider.get_pkce_params().keys()) == pkce_enabled



from allauth.headless.base.response import APIResponse
from allauth.usersessions import app_settings
from django.db import connection
from zango.core.utils import get_datetime_str_in_tenant_timezone

class SessionsResponse(APIResponse):
    def __init__(self, request, sessions):
        super().__init__(request, data=[self._session_data(s) for s in sessions])

    def _session_data(self, session):
        data = {
            "user_agent": session.user_agent,
            "ip": session.ip,
            "created_at": get_datetime_str_in_tenant_timezone(session.created_at, connection.tenant),
            "is_current": session.is_current(),
            "id": session.pk,
        }
        # if app_settings.TRACK_ACTIVITY:
        data["last_seen_at"] = get_datetime_str_in_tenant_timezone(session.last_seen_at, connection.tenant)
        return data


def get_config_data(request):
    data = {"usersessions": {"track_activity": app_settings.TRACK_ACTIVITY}}
    return data



from django.urls import include, path

from allauth.idp.oidc import views


app_name = "oidc"
urlpatterns = [
    path(
        ".well-known/",
        include(
            [
                path(
                    "openid-configuration",
                    views.configuration,
                    name="configuration",
                ),
                path(
                    "jwks.json",
                    views.jwks,
                    name="jwks",
                ),
            ]
        ),
    ),
    path(
        "identity/",
        include(
            [
                path(
                    "o/",
                    include(
                        [
                            path(
                                "authorize",
                                views.authorization,
                                name="authorization",
                            ),
                            path(
                                "device",
                                views.device_authorization,
                                name="device_authorization",
                            ),
                            path(
                                "api/",
                                include(
                                    [
                                        path(
                                            "token",
                                            views.token,
                                            name="token",
                                        ),
                                        path(
                                            "revoke",
                                            views.revoke,
                                            name="revoke",
                                        ),
                                        path(
                                            "userinfo",
                                            views.user_info,
                                            name="userinfo",
                                        ),
                                        path(
                                            "device/code",
                                            views.device_code,
                                            name="device_code",
                                        ),
                                    ]
                                ),
                            ),
                        ]
                    ),
                )
            ]
        ),
    ),
]



from allauth.socialaccount import app_settings
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.app_settings import QUERY_EMAIL
from allauth.socialaccount.providers.oauth2.views import (
    OAuth2Adapter,
    OAuth2CallbackView,
    OAuth2LoginView,
)


class LichessOAuth2Adapter(OAuth2Adapter):
    provider_id = "lichess"

    settings = app_settings.PROVIDERS.get(provider_id, {})
    provider_base_url = settings.get("API_URL", "https://lichess.org")

    access_token_url = f"{provider_base_url}/api/token"
    authorize_url = f"{provider_base_url}/oauth"

    profile_url = f"{provider_base_url}/api/account"
    email_address_url = f"{provider_base_url}/api/account/email"

    def complete_login(self, request, app, token, response):
        profile_res = (
            get_adapter()
            .get_requests_session()
            .get(
                self.profile_url,
                params={"access_token": token.token},
                headers={"Authorization": f"Bearer {token.token}"},
            )
        )

        profile_res.raise_for_status()
        extra_data = profile_res.json()

        user_profile = extra_data["result"] if "result" in extra_data else extra_data

        # retrieve email address if requested
        if QUERY_EMAIL:
            email_data = (
                get_adapter()
                .get_requests_session()
                .get(
                    self.email_address_url,
                    headers={"Authorization": f"Bearer {token.token}"},
                )
            )

            email_data.raise_for_status()
            email_data = email_data.json()

            # extract email address from response

            email = email_data.get("email", None)

            if email:
                user_profile["email"] = email

        return self.get_provider().sociallogin_from_response(request, user_profile)


oauth2_login = OAuth2LoginView.adapter_view(LichessOAuth2Adapter)
oauth2_callback = OAuth2CallbackView.adapter_view(LichessOAuth2Adapter)



import typing

from django.http import HttpRequest

from rest_framework import authentication

from allauth.headless.internal.sessionkit import authenticate_by_x_session_token


class XSessionTokenAuthentication(authentication.BaseAuthentication):
    """
    This authentication class uses the X-Session-Token that django-allauth
    is using for authentication purposes.
    """

    def authenticate(self, request: HttpRequest):
        token = self.get_session_token(request)
        if token:
            return authenticate_by_x_session_token(token)
        return None

    def get_session_token(self, request: HttpRequest) -> typing.Optional[str]:
        """
        Returns the session token for the given request, by looking up the
        ``X-Session-Token`` header. Override this if you want to extract the token
        from e.g. the ``Authorization`` header.
        """
        return request.headers.get("X-Session-Token")



import json

from django.test import TestCase

from allauth.socialaccount.providers.microsoft.provider import MicrosoftGraphProvider
from allauth.socialaccount.providers.microsoft.views import _check_errors
from allauth.socialaccount.providers.oauth2.client import OAuth2Error
from tests.apps.socialaccount.base import OAuth2TestsMixin
from tests.mocking import MockedResponse


class MicrosoftGraphTests(OAuth2TestsMixin, TestCase):
    provider_id = MicrosoftGraphProvider.id

    def get_mocked_response(self):
        response_data = """
        {
            "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users/$entity",
            "id": "16f5a7b6-5a15-4568-aa5a-31bb117e9967",
            "businessPhones": [],
            "displayName": "Anne Weiler",
            "givenName": "Anne",
            "jobTitle": "Manufacturing Lead",
            "mail": "annew@CIE493742.onmicrosoft.com",
            "mobilePhone": "+1 3528700812",
            "officeLocation": null,
            "preferredLanguage": "en-US",
            "surname": "Weiler",
            "userPrincipalName": "annew@CIE493742.onmicrosoft.com",
            "mailNickname": "annew"
        }
        """  # noqa
        return MockedResponse(200, response_data)

    def get_expected_to_str(self):
        return "annew@CIE493742.onmicrosoft.com"

    def test_invalid_data(self):
        response = MockedResponse(200, json.dumps({}))
        with self.assertRaises(OAuth2Error):
            # No id, raises
            _check_errors(response)

    def test_profile_invalid_response(self):
        data = {
            "error": {
                "code": "InvalidAuthenticationToken",
                "message": "Access token validation failure. Invalid audience.",
            }
        }
        response = MockedResponse(401, json.dumps(data))

        with self.assertRaises(OAuth2Error):
            # no id, 4xx code, raises with message
            _check_errors(response)

    def test_invalid_response(self):
        response = MockedResponse(200, "invalid json data")
        with self.assertRaises(OAuth2Error):
            # bad json, raises
            _check_errors(response)



import os

import nox


@nox.session
def docs(session):
    session.install(
        "Django==5.2",
        "Sphinx>=7.4.7,<8",
        "sphinx_rtd_theme",
        "djangorestframework>=3.15.2,<4",
        "django-ninja>=1.3.0,<2",
        "snowballstemmer<3",  # https://github.com/sphinx-doc/sphinx/issues/13533
    )
    session.run("make", "docs")


@nox.session(tags=["lint"])
def lint(session):
    session.install("bandit")
    session.run("bandit", "-c", "pyproject.toml", "-r", "allauth/")


@nox.session(tags=["lint"])
def isort(session):
    session.install("isort==5.13.2")
    session.run("isort", "--check-only", "--diff", "--gitignore", ".")


@nox.session(tags=["lint"])
def flake8(session):
    session.install("flake8==7.2.0")
    session.run("flake8", "allauth/")


@nox.session(tags=["lint"])
def black(session):
    session.install("black==24.4.0")
    session.run("black", "--check", ".")


@nox.session(tags=["lint"])
def djlint(session):
    session.install("djlint==1.34.1")
    session.run(
        "djlint", "--configuration", ".djlintrc", "--check", "./allauth", "./examples"
    )


DJANGO_PYTHON_REQ = {
    "4.2.20": ("3.8", "3.9", "3.10", "3.11"),
    "5.0": ("3.10", "3.11", "3.12"),
    "5.1": ("3.10", "3.11", "3.12"),
    "5.2": ("3.10", "3.11", "3.12", "3.13"),
}
DJANGO_LTS = "5.2"


@nox.session(python=["3.8", "3.9", "3.10", "3.11", "3.12", "3.13"])
@nox.parametrize("django", list(DJANGO_PYTHON_REQ.keys()))
@nox.parametrize(
    "project", ["regular", "headless_only", "account_only", "login_required_mw"]
)
def test(session, django, project):
    django_version = tuple(map(int, django.split(".")))
    if django != DJANGO_LTS:
        last_django_python = DJANGO_PYTHON_REQ[django][-1]
        first_django_python = DJANGO_PYTHON_REQ[django][0]
        if session.python not in (first_django_python, last_django_python):
            print(
                f"Skipping: only current LTS is tested against python{session.python}, not Django {django}"
            )
            return

    if project == "login_required_mw" and django_version < (5, 1):
        print(f"Skipping: project {project} does not support Django {django}")
        return
    if session.python not in DJANGO_PYTHON_REQ[django]:
        print(f"Skipping: Django {django} does not support python{session.python}")
        return
    session.install(
        f"django=={django}",
        "pytest>=8.3.5,<9",
        "pytest-asyncio==0.23.8",
        "pytest-django>=4.11,<5",
        "Pillow>=9.0",
        "coverage==7.6.1",
        # SAML is disabled in CI
        # "python3-saml==1.16.0",
        # "xmlsec==1.3.14",
        # "lxml==5.2.1",
        "pyyaml>=6.0.2,<7",
        "psycopg2-binary>=2.9.10,<3",
        "djangorestframework>=3.15.2,<4",
        "django-ninja>=1.3.0,<2",
        "mypy==1.10.0",
        ".[mfa,openid,socialaccount,steam]",  # SAML is disabled in CI
    )
    session.run("/bin/sh", "-c", "cd allauth; python ../manage.py compilemessages")
    run_coveralls = (
        os.environ.get("GITHUB_TOKEN")
        and project == "regular"
        and django == "5.2"
        and session.python == "3.13"
    )
    if run_coveralls:
        session.install("coveralls")
    session.run(
        "coverage",
        "run",
        "-m",
        "pytest",
        f"--ds=tests.projects.{project}.settings",
        "tests/",
    )
    if run_coveralls:
        session.run("coveralls", "--service=github")
    if django == "5.1" and session.python == "3.13":
        session.install(
            "django-stubs==5.1.3",
            "types-requests==2.32.0.20240602",
        )
        session.run("mypy", "allauth")



from django.conf import settings
from django.contrib.auth import get_user_model
from django.urls import reverse

from allauth.socialaccount.models import SocialAccount


def test_login(client, db):
    resp = client.post(reverse("dummy_login"))
    assert resp.status_code == 302
    assert resp["location"].startswith(reverse("dummy_authenticate") + "?state=")
    resp = client.post(
        resp["location"],
        {"id": "123", "email": "a@b.com", "email_verified": True},
    )
    assert resp.status_code == 302
    assert resp["location"] == settings.LOGIN_REDIRECT_URL
    get_user_model().objects.filter(email="a@b.com").exists()
    socialaccount = SocialAccount.objects.get(uid="123")
    account = socialaccount.get_provider_account()
    assert account.to_str() == "a@b.com"



import urllib.error
from unittest.mock import Mock, patch

from django.contrib.auth import get_user_model
from django.test import override_settings
from django.urls import reverse

import pytest
from openid import fetchers
from openid.consumer import consumer

from allauth.socialaccount.models import SocialAccount
from allauth.socialaccount.providers.openid import views
from allauth.socialaccount.providers.openid.utils import AXAttribute


class TestFetcher(fetchers.Urllib2Fetcher):
    def fetch(self, url, body=None, headers=None):
        if url == "https://steamcommunity.com/openid":
            return fetchers.HTTPResponse(
                final_url="https://steamcommunity.com/openid",
                status=200,
                headers={"content-type": "application/xrds+xml;charset=utf-8"},
                body='<?xml version="1.0" encoding="UTF-8"?>\n<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">\n\t<XRD>\n\t\t<Service priority="0">\n\t\t\t<Type>http://specs.openid.net/auth/2.0/server</Type>\t\t\n\t\t\t<URI>https://steamcommunity.com/openid/login</URI>\n\t\t</Service>\n\t</XRD>\n</xrds:XRDS>',
            )
        if url == "https://steamcommunity.com/openid/login":
            return fetchers.HTTPResponse(
                final_url="https://steamcommunity.com/openid/login",
                status=200,
                headers={"content-type": "text/plain;charset=utf-8"},
                body="ns:http://specs.openid.net/auth/2.0\nerror_code:unsupported-type\nerror:Associations not supported\n",
            )

        if url == "https://discovery-failure.com/":
            raise urllib.error.URLError
        ret = super().fetch(url, body=body, headers=headers)
        breakpoint()
        return ret


@pytest.fixture(autouse=True)
def setup_fetcher():
    old_fetcher = fetchers.getDefaultFetcher()
    fetchers.setDefaultFetcher(TestFetcher())
    yield
    fetchers.setDefaultFetcher(old_fetcher)


def test_discovery_failure(client):
    """
    This used to generate a server 500:
    DiscoveryFailure: No usable OpenID services found
    for http://www.google.com/
    """
    resp = client.post(
        reverse("openid_login"), dict(openid="https://discovery-failure.com/")
    )
    assert "openid" in resp.context["form"].errors


def test_login(client, db):
    # Location: https://s.yimg.com/wm/mbr/html/openid-eol-0.0.1.html
    resp = client.post(
        reverse(views.login), dict(openid="https://steamcommunity.com/openid")
    )
    assert "steamcommunity.com/openid/login" in resp["location"]
    with patch(
        "allauth.socialaccount.providers.openid.views._openid_consumer"
    ) as consumer_mock:
        consumer_client = Mock()
        complete = Mock()
        consumer_mock.return_value = consumer_client
        consumer_client.complete = complete
        complete_response = Mock()
        complete.return_value = complete_response
        complete_response.status = consumer.SUCCESS
        complete_response.identity_url = "http://dummy/john/"
        with patch(
            "allauth.socialaccount.providers.openid.utils.SRegResponse"
        ) as sr_mock:
            with patch(
                "allauth.socialaccount.providers.openid.utils.FetchResponse"
            ) as fr_mock:
                sreg_mock = Mock()
                ax_mock = Mock()
                sr_mock.fromSuccessResponse = sreg_mock
                fr_mock.fromSuccessResponse = ax_mock
                sreg_mock.return_value = {}
                ax_mock.return_value = {AXAttribute.PERSON_FIRST_NAME: ["raymond"]}
                resp = client.post(reverse("openid_callback"))
                assert resp["location"] == "/accounts/profile/"
                get_user_model().objects.get(first_name="raymond")
                social_account = SocialAccount.objects.get(
                    uid=complete_response.identity_url
                )
                account = social_account.get_provider_account()
                assert account.to_str() == complete_response.identity_url


@override_settings(
    SOCIALACCOUNT_PROVIDERS={
        "openid": {
            "SERVERS": [
                dict(
                    id="yahoo",
                    name="Yahoo",
                    openid_url="http://me.yahoo.com",
                    extra_attributes=[
                        (
                            "phone",
                            "http://axschema.org/contact/phone/default",
                            True,
                        )
                    ],
                )
            ]
        }
    }
)
def test_login_with_extra_attributes(client, db):
    with patch("allauth.socialaccount.providers.openid.views.QUERY_EMAIL", True):
        resp = client.post(
            reverse(views.login), dict(openid="https://steamcommunity.com/openid")
        )
    assert "steamcommunity.com/openid/login" in resp["location"]
    with patch(
        "allauth.socialaccount.providers.openid.views._openid_consumer"
    ) as consumer_mock:
        consumer_client = Mock()
        complete = Mock()
        endpoint = Mock()
        consumer_mock.return_value = consumer_client
        consumer_client.complete = complete
        complete_response = Mock()
        complete.return_value = complete_response
        complete_response.endpoint = endpoint
        complete_response.endpoint.server_url = "http://me.yahoo.com"
        complete_response.status = consumer.SUCCESS
        complete_response.identity_url = "http://dummy/john/"
        with patch(
            "allauth.socialaccount.providers.openid.utils.SRegResponse"
        ) as sr_mock:
            with patch(
                "allauth.socialaccount.providers.openid.utils.FetchResponse"
            ) as fr_mock:
                sreg_mock = Mock()
                ax_mock = Mock()
                sr_mock.fromSuccessResponse = sreg_mock
                fr_mock.fromSuccessResponse = ax_mock
                sreg_mock.return_value = {}
                ax_mock.return_value = {
                    AXAttribute.CONTACT_EMAIL: ["raymond@example.com"],
                    AXAttribute.PERSON_FIRST_NAME: ["raymond"],
                    "http://axschema.org/contact/phone/default": ["123456789"],
                }
                resp = client.post(reverse("openid_callback"))
                assert resp["location"] == "/accounts/profile/"
                socialaccount = SocialAccount.objects.get(user__first_name="raymond")
                assert socialaccount.extra_data.get("phone") == "123456789"



from typing import Optional, Tuple


class AppSettings:
    def __init__(self, prefix):
        self.prefix = prefix

    def _setting(self, name, dflt):
        from allauth.utils import get_setting

        return get_setting(self.prefix + name, dflt)

    @property
    def ADAPTER(self):
        return self._setting(
            "ADAPTER", "allauth.headless.adapter.DefaultHeadlessAdapter"
        )

    @property
    def TOKEN_STRATEGY(self):
        from allauth.utils import import_attribute

        path = self._setting(
            "TOKEN_STRATEGY", "allauth.headless.tokens.sessions.SessionTokenStrategy"
        )
        cls = import_attribute(path)
        return cls()

    @property
    def SERVE_SPECIFICATION(self) -> bool:
        return self._setting("SERVE_SPECIFICATION", False)

    @property
    def SPECIFICATION_TEMPLATE_NAME(self) -> Optional[str]:
        return self._setting(
            "SPECIFICATION_TEMPLATE_NAME", "headless/spec/redoc_cdn.html"
        )

    @property
    def CLIENTS(self) -> Tuple[str]:
        return tuple(self._setting("CLIENTS", ("browser", "app")))

    @property
    def FRONTEND_URLS(self):
        return self._setting("FRONTEND_URLS", {})


_app_settings = AppSettings("HEADLESS_")


def __getattr__(name):
    # See https://peps.python.org/pep-0562/
    return getattr(_app_settings, name)



import time
from contextlib import contextmanager
from unittest.mock import patch

from django.urls import reverse

import pytest

from allauth.account.models import EmailAddress
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.internal import statekit
from allauth.socialaccount.models import SocialAccount, SocialLogin, SocialToken
from allauth.socialaccount.providers.base.constants import AuthProcess
from tests.mocking import MockedResponse, mocked_response


@pytest.fixture
def sociallogin_factory(user_factory):
    def factory(
        email=None,
        username=None,
        with_email=True,
        provider="unittest-server",
        uid="123",
        email_verified=True,
        with_token=False,
    ):
        user = user_factory(
            username=username, email=email, commit=False, with_email=with_email
        )
        provider_instance = get_adapter().get_provider(request=None, provider=provider)
        account = SocialAccount(provider=provider, uid=uid)
        sociallogin = SocialLogin(
            provider=provider_instance, user=user, account=account
        )
        if with_email:
            sociallogin.email_addresses = [
                EmailAddress(email=user.email, verified=email_verified, primary=True)
            ]
        if with_token:
            sociallogin.token = SocialToken(token="123", token_secret="456")  # nosec
        return sociallogin

    return factory


@pytest.fixture
def jwt_decode_bypass():
    @contextmanager
    def f(jwt_data):
        with patch("allauth.socialaccount.internal.jwtkit.verify_and_decode") as m:
            data = {
                "iss": "https://accounts.google.com",
                "aud": "client_id",
                "sub": "123sub",
                "hd": "example.com",
                "email": "raymond@example.com",
                "email_verified": True,
                "at_hash": "HK6E_P6Dh8Y93mRNtsDB1Q",
                "name": "Raymond Penners",
                "picture": "https://lh5.googleusercontent.com/photo.jpg",
                "given_name": "Raymond",
                "family_name": "Penners",
                "locale": "en",
                "iat": 123,
                "exp": 456,
            }
            data.update(jwt_data)
            m.return_value = data
            yield

    return f


@pytest.fixture
def provider_callback_response():
    def f(client, process=AuthProcess.LOGIN):
        with mocked_response(
            {
                "token_endpoint": "/",
                "userinfo_endpoint": "/",
            },
            MockedResponse(200, "access_token=456", {"content-type": "dummy"}),
            {
                "sub": "sub123",
            },
        ):
            session = client.session
            session[statekit.STATES_SESSION_KEY] = {
                "state456": [{"process": process}, time.time()]
            }
            session.save()

            resp = client.post(
                reverse(
                    "openid_connect_callback",
                    kwargs={"provider_id": "unittest-server"},
                )
                + "?code=123&state=state456"
            )
            return resp

    return f



from unittest.mock import ANY

from django.conf import settings
from django.contrib.auth import get_user_model
from django.urls import reverse

import pytest
from pytest_django.asserts import assertTemplateUsed

from allauth.account.authentication import AUTHENTICATION_METHODS_SESSION_KEY
from allauth.mfa.models import Authenticator


def test_passkey_login(client, passkey, webauthn_authentication_bypass):
    with webauthn_authentication_bypass(passkey) as credential:
        resp = client.get(
            reverse("mfa_login_webauthn"), HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        assert "request_options" in resp.json()
        resp = client.post(
            reverse("mfa_login_webauthn"), data={"credential": credential}
        )
    assert resp["location"] == settings.LOGIN_REDIRECT_URL
    assert client.session[AUTHENTICATION_METHODS_SESSION_KEY] == [
        {
            "at": ANY,
            "id": ANY,
            "method": "mfa",
            "passwordless": True,
            "type": "webauthn",
        }
    ]


def test_reauthenticate(
    auth_client, passkey, user_with_recovery_codes, webauthn_authentication_bypass
):
    resp = auth_client.get(reverse("mfa_view_recovery_codes"))
    assert resp.status_code == 302
    assert resp["location"].startswith(reverse("account_reauthenticate"))
    resp = auth_client.get(reverse("mfa_reauthenticate"))
    assertTemplateUsed(resp, "mfa/reauthenticate.html")

    with webauthn_authentication_bypass(passkey) as credential:
        resp = auth_client.get(
            reverse("mfa_reauthenticate_webauthn"),
        )
        resp = auth_client.post(
            reverse("mfa_reauthenticate_webauthn"),
            data={"credential": credential, "next": "/redir"},
        )
    assert resp["location"] == "/redir"


def test_get_passkey_login_challenge_redirects_if_not_ajax(client):
    resp = client.get(reverse("mfa_login_webauthn"))
    assert resp["location"] == reverse("account_login")


def test_get_passkey_login_challenge(client, db):
    resp = client.get(
        reverse("mfa_login_webauthn"), HTTP_X_REQUESTED_WITH="XMLHttpRequest"
    )
    assert resp.status_code == 200
    assert resp["content-type"] == "application/json"
    data = resp.json()
    assert data == {
        "request_options": {
            "publicKey": {
                "challenge": ANY,
                "rpId": "testserver",
                "allowCredentials": [],
                "userVerification": "preferred",
            }
        }
    }


def test_invalid_passkey_login(client, passkey):
    resp = client.post(reverse("mfa_login_webauthn"), data={"credential": "{}"})
    assert resp["location"] == reverse("account_login")


def test_rename_key(auth_client, passkey, reauthentication_bypass):
    resp = auth_client.get(reverse("mfa_edit_webauthn", kwargs={"pk": passkey.pk}))
    assert resp["location"].startswith(reverse("account_reauthenticate"))
    with reauthentication_bypass():
        resp = auth_client.get(reverse("mfa_edit_webauthn", kwargs={"pk": passkey.pk}))
        assertTemplateUsed(resp, "mfa/webauthn/edit_form.html")
        resp = auth_client.post(
            reverse("mfa_edit_webauthn", kwargs={"pk": passkey.pk}),
            data={"name": "Renamed"},
        )
        assert resp["location"] == reverse("mfa_list_webauthn")
        passkey.refresh_from_db()
        assert passkey.data["name"] == "Renamed"
        assert str(passkey) == "Renamed"


def test_remove_key(auth_client, passkey, reauthentication_bypass):
    resp = auth_client.get(reverse("mfa_remove_webauthn", kwargs={"pk": passkey.pk}))
    assert resp["location"].startswith(reverse("account_reauthenticate"))
    with reauthentication_bypass():
        resp = auth_client.get(
            reverse("mfa_remove_webauthn", kwargs={"pk": passkey.pk})
        )
        assertTemplateUsed(resp, "mfa/webauthn/authenticator_confirm_delete.html")
        resp = auth_client.post(
            reverse("mfa_remove_webauthn", kwargs={"pk": passkey.pk})
        )
        assert resp["location"] == reverse("mfa_list_webauthn")


@pytest.mark.parametrize("passwordless", [False, True])
def test_add_key(
    auth_client,
    user,
    webauthn_registration_bypass,
    reauthentication_bypass,
    passwordless,
):
    with webauthn_registration_bypass(user, passwordless) as credential:
        resp = auth_client.post(
            reverse("mfa_add_webauthn"), data={"credential": credential}
        )
        assert resp["location"].startswith(reverse("account_reauthenticate"))
    with reauthentication_bypass():
        resp = auth_client.get(reverse("mfa_add_webauthn"))
        assertTemplateUsed(resp, "mfa/webauthn/add_form.html")
        with webauthn_registration_bypass(user, passwordless) as credential:
            resp = auth_client.post(
                reverse("mfa_add_webauthn"),
                data={
                    "credential": credential,
                    "passwordless": "on" if passwordless else "",
                },
            )
            assert resp["location"].startswith(reverse("mfa_view_recovery_codes"))
        authenticator = Authenticator.objects.get(
            user=user, type=Authenticator.Type.WEBAUTHN
        )
        assert authenticator.wrap().is_passwordless == passwordless
        Authenticator.objects.filter(
            user=user, type=Authenticator.Type.RECOVERY_CODES
        ).exists()


def test_list_keys(auth_client):
    resp = auth_client.get(reverse("mfa_list_webauthn"))
    assertTemplateUsed(resp, "mfa/webauthn/authenticator_list.html")


@pytest.mark.parametrize("email_verified", [False])
@pytest.mark.parametrize("method", ["get", "post"])
@pytest.mark.parametrize("allowed", [True, False])
def test_add_with_unverified_email(
    auth_client,
    user,
    webauthn_registration_bypass,
    reauthentication_bypass,
    method,
    allowed,
    settings,
):
    settings.MFA_ALLOW_UNVERIFIED_EMAIL = allowed
    with webauthn_registration_bypass(user, False) as credential:
        if method == "get":
            resp = auth_client.get(reverse("mfa_add_webauthn"))
        else:
            resp = auth_client.post(
                reverse("mfa_add_webauthn"), data={"credential": credential}
            )
        if allowed:
            assert resp["location"].startswith(reverse("account_reauthenticate"))
        else:
            assert resp["location"] == reverse("mfa_index")


def test_passkey_signup(client, db, webauthn_registration_bypass):
    resp = client.post(
        reverse("account_signup_by_passkey"),
        data={"email": "pass@key.org", "username": "passkey"},
    )
    assert resp["location"] == reverse("mfa_signup_webauthn")
    resp = client.post(resp["location"])
    assert resp.status_code == 200
    user = get_user_model().objects.get(email="pass@key.org")
    with webauthn_registration_bypass(user, True) as credential:
        resp = client.post(
            reverse("mfa_signup_webauthn"), data={"credential": credential}
        )
    assert resp["location"] == settings.LOGIN_REDIRECT_URL


def test_webauthn_login(
    client, user_with_passkey, passkey, user_password, webauthn_authentication_bypass
):
    resp = client.post(
        reverse("account_login"),
        {"login": user_with_passkey.username, "password": user_password},
    )
    assert resp.status_code == 302
    assert resp["location"] == reverse("mfa_authenticate")
    with webauthn_authentication_bypass(passkey) as credential:
        resp = client.get(reverse("mfa_authenticate"))
        assert resp.status_code == 200
        resp = client.post(reverse("mfa_authenticate"), {"credential": credential})
    assert resp.status_code == 302
    assert resp["location"] == settings.LOGIN_REDIRECT_URL
    assert client.session[AUTHENTICATION_METHODS_SESSION_KEY] == [
        {"method": "password", "at": ANY, "username": user_with_passkey.username},
        {"method": "mfa", "at": ANY, "id": ANY, "type": Authenticator.Type.WEBAUTHN},
    ]



from allauth.mfa.adapter import get_adapter


def encrypt(text):
    return get_adapter().encrypt(text)


def decrypt(encrypted_text):
    return get_adapter().decrypt(encrypted_text)


def is_mfa_enabled(user=None, request=None):
    return get_adapter().is_mfa_enabled(user, request)



# Generated by Django 4.2.16 on 2025-07-08 05:02

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('appauth', '0007_appuserauthtoken'),
        ('account', '0009_emailaddress_unique_primary_email'),
    ]

    operations = [
        migrations.AlterField(
            model_name='emailaddress',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='appauth.appusermodel', verbose_name='user'),
        ),
    ]



from allauth.mfa.models import Authenticator


def test_get_recovery_codes_requires_reauth(
    auth_client, user_with_recovery_codes, headless_reverse
):
    rc = Authenticator.objects.get(
        type=Authenticator.Type.RECOVERY_CODES, user=user_with_recovery_codes
    )
    resp = auth_client.get(headless_reverse("headless:mfa:manage_recovery_codes"))
    assert resp.status_code == 401
    data = resp.json()
    assert data["meta"]["is_authenticated"]
    resp = auth_client.post(
        headless_reverse("headless:mfa:reauthenticate"),
        data={"code": rc.wrap().get_unused_codes()[0]},
        content_type="application/json",
    )
    assert resp.status_code == 200


def test_get_recovery_codes(
    auth_client,
    user_with_recovery_codes,
    headless_reverse,
    reauthentication_bypass,
):
    with reauthentication_bypass():
        resp = auth_client.get(headless_reverse("headless:mfa:manage_recovery_codes"))
    assert resp.status_code == 200
    data = resp.json()
    assert data["data"]["type"] == "recovery_codes"
    assert len(data["data"]["unused_codes"]) == 10

    with reauthentication_bypass():
        resp = auth_client.get(headless_reverse("headless:mfa:authenticators"))
    data = resp.json()
    assert len(data["data"]) == 2
    rc = [autor for autor in data["data"] if autor["type"] == "recovery_codes"][0]
    assert "unused_codes" not in rc


def test_generate_recovery_codes(
    auth_client,
    user_with_totp,
    headless_reverse,
    reauthentication_bypass,
):
    with reauthentication_bypass():
        resp = auth_client.get(headless_reverse("headless:mfa:manage_recovery_codes"))
    assert resp.status_code == 404
    with reauthentication_bypass():
        resp = auth_client.post(
            headless_reverse("headless:mfa:manage_recovery_codes"),
            content_type="application/json",
        )
    assert resp.status_code == 200
    data = resp.json()
    assert data["data"]["type"] == "recovery_codes"
    assert len(data["data"]["unused_codes"]) == 10



from unittest.mock import patch

from django.test import TestCase

from allauth.account.models import EmailAddress
from allauth.socialaccount.models import SocialAccount
from allauth.socialaccount.providers.github.provider import GitHubProvider
from tests.apps.socialaccount.base import OAuth2TestsMixin
from tests.mocking import MockedResponse


class GitHubTests(OAuth2TestsMixin, TestCase):
    provider_id = GitHubProvider.id

    def get_mocked_response(self):
        return [
            MockedResponse(
                200,
                """
        {
            "type":"User",
            "organizations_url":"https://api.github.com/users/pennersr/orgs",
            "gists_url":"https://api.github.com/users/pennersr/gists{/gist_id}",
            "received_events_url":"https://api.github.com/users/pennersr/received_events",
            "gravatar_id":"8639768262b8484f6a3380f8db2efa5b",
            "followers":16,
            "blog":"http://www.intenct.info",
            "avatar_url":"https://secure.gravatar.com/avatar/8639768262b8484f6a3380f8db2efa5b?d=https://a248.e.akamai.net/assets.github.com%2Fimages%2Fgravatars%2Fgravatar-user-420.png",
            "login":"pennersr",
            "created_at":"2010-02-10T12:50:51Z",
            "company":"IntenCT",
            "subscriptions_url":"https://api.github.com/users/pennersr/subscriptions",
            "public_repos":14,
            "hireable":false,
            "url":"https://api.github.com/users/pennersr",
            "public_gists":0,
            "starred_url":"https://api.github.com/users/pennersr/starred{/owner}{/repo}",
            "html_url":"https://github.com/pennersr",
            "location":"The Netherlands",
            "bio":null,
            "name":"Raymond Penners",
            "repos_url":"https://api.github.com/users/pennersr/repos",
            "followers_url":"https://api.github.com/users/pennersr/followers",
            "id":201022,
            "following":0,
            "email":"raymond.penners@intenct.nl",
            "events_url":"https://api.github.com/users/pennersr/events{/privacy}",
            "following_url":"https://api.github.com/users/pennersr/following"
        }""",
            ),
            MockedResponse(
                200,
                """
            [{
              "email": "octocat@github.com",
              "verified": true,
              "primary": true,
              "visibility": "public"
            }]
            """,
            ),
        ]

    def get_expected_to_str(self):
        return "pennersr"

    def test_account_name_null(self):
        """String conversion when GitHub responds with empty name"""
        mocks = [
            MockedResponse(
                200,
                """
        {
            "type": "User",
            "id": 201022,
            "login": "pennersr",
            "name": null
        }
        """,
            ),
            MockedResponse(
                200,
                """
        [
          {
            "email": "octocat@github.com",
            "verified": true,
            "primary": true,
            "visibility": "public"
          },
          {
            "email": "secONDary@GitHub.COM",
            "verified": true,
            "primary": false,
            "visibility": "public"
          }
        ]
        """,
            ),
        ]
        with patch(
            "allauth.socialaccount.adapter.DefaultSocialAccountAdapter.populate_user"
        ) as populate_mock:
            self.login(mocks)
        populate_data = populate_mock.call_args[0][2]
        assert populate_data["email"] == "octocat@github.com"
        socialaccount = SocialAccount.objects.get(uid="201022")
        self.assertIsNone(socialaccount.extra_data.get("name"))
        account = socialaccount.get_provider_account()
        self.assertIsNotNone(account.to_str())
        self.assertEqual(account.to_str(), "pennersr")
        self.assertEqual(socialaccount.user.email, "octocat@github.com")
        self.assertTrue(
            EmailAddress.objects.filter(
                primary=False,
                verified=True,
                email="secondary@github.com",
                user=socialaccount.user,
            ).exists()
        )
        self.assertTrue("emails" not in socialaccount.extra_data)



from allauth.account.models import EmailAddress, get_emailconfirmation_model
from allauth.headless.constants import Flow


def test_auth_unverified_email_and_mfa(
    client,
    user_factory,
    password_factory,
    settings,
    totp_validation_bypass,
    headless_reverse,
    headless_client,
):
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    settings.ACCOUNT_EMAIL_VERIFICATION = "mandatory"
    settings.ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION = True
    password = password_factory()
    user = user_factory(email_verified=False, password=password, with_totp=True)
    resp = client.post(
        headless_reverse("headless:account:login"),
        data={
            "email": user.email,
            "password": password,
        },
        content_type="application/json",
    )
    assert resp.status_code == 401
    data = resp.json()
    assert [f for f in data["data"]["flows"] if f["id"] == Flow.VERIFY_EMAIL][0][
        "is_pending"
    ]
    emailaddress = EmailAddress.objects.filter(user=user, verified=False).get()
    key = get_emailconfirmation_model().create(emailaddress).key
    resp = client.post(
        headless_reverse("headless:account:verify_email"),
        data={"key": key},
        content_type="application/json",
    )
    assert resp.status_code == 401
    flows = [
        {"id": "login"},
        {"id": "login_by_code"},
        {"id": "signup"},
    ]
    if headless_client == "browser":
        flows.append(
            {
                "id": "provider_redirect",
                "providers": ["dummy", "unittest-server", "other-server"],
            }
        )
    flows.append(
        {
            "id": "provider_token",
            "providers": ["dummy", "unittest-server", "other-server"],
        }
    )
    flows.append({"id": "mfa_login_webauthn"})
    flows.append(
        {
            "id": "mfa_authenticate",
            "is_pending": True,
            "types": ["totp"],
        }
    )

    assert resp.json() == {
        "data": {"flows": flows},
        "meta": {"is_authenticated": False},
        "status": 401,
    }
    resp = client.post(
        headless_reverse("headless:mfa:authenticate"),
        data={"code": "bad"},
        content_type="application/json",
    )
    assert resp.status_code == 400
    assert resp.json() == {
        "status": 400,
        "errors": [
            {"message": "Incorrect code.", "code": "incorrect_code", "param": "code"}
        ],
    }

    with totp_validation_bypass():
        resp = client.post(
            headless_reverse("headless:mfa:authenticate"),
            data={"code": "bad"},
            content_type="application/json",
        )
    assert resp.status_code == 200


def test_dangling_mfa_is_logged_out(
    client,
    user_with_totp,
    password_factory,
    settings,
    totp_validation_bypass,
    headless_reverse,
    headless_client,
    user_password,
):
    settings.ACCOUNT_LOGIN_METHODS = {"email"}
    resp = client.post(
        headless_reverse("headless:account:login"),
        data={
            "email": user_with_totp.email,
            "password": user_password,
        },
        content_type="application/json",
    )
    assert resp.status_code == 401
    data = resp.json()
    flow = [f for f in data["data"]["flows"] if f["id"] == Flow.MFA_AUTHENTICATE][0]
    assert flow["is_pending"]
    assert flow["types"] == ["totp"]
    resp = client.delete(headless_reverse("headless:account:current_session"))
    data = resp.json()
    assert resp.status_code == 401
    assert all(not f.get("is_pending") for f in data["data"]["flows"])



SECRET_KEY = "psst"
USE_I18N = False
USE_TZ = True


MIDDLEWARE = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "allauth.account.middleware.AccountMiddleware",
)

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.admin",
    "django.contrib.humanize",
    "allauth",
    "allauth.account",
    "allauth.mfa",
    "allauth.socialaccount",
    "allauth.usersessions",
    "allauth.headless",
)

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",
    "allauth.account.auth_backends.AuthenticationBackend",
)



from unittest.mock import Mock, patch
from urllib.parse import parse_qs, urlparse

from django.conf import settings
from django.urls import reverse, reverse_lazy
from django.utils.http import urlencode

import pytest
from pytest_django.asserts import assertTemplateUsed

from allauth.account.models import EmailAddress
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.internal import statekit
from allauth.socialaccount.models import SocialAccount
from allauth.socialaccount.providers.base.constants import AuthProcess
from allauth.socialaccount.providers.saml.utils import build_saml_config


@pytest.mark.parametrize(
    "idp_initiated,adv_settings,state_kwargs,relay_state, expected_url",
    [
        (False, {}, {}, "/not/here", settings.LOGIN_REDIRECT_URL),
        (False, {}, {"next": "/here"}, "/not/here", "/here"),
        (
            False,
            {},
            {"process": "connect"},
            "/not/here",
            reverse_lazy("socialaccount_connections"),
        ),
        (False, {}, {"process": "connect", "next": "/here"}, "/not/here", "/here"),
        (True, {"reject_idp_initiated_sso": False}, {}, "/set-by-idp", "/set-by-idp"),
        (
            True,
            {"reject_idp_initiated_sso": False},
            {},
            "not-a-url",
            settings.LOGIN_REDIRECT_URL,
        ),
        (True, {}, {}, "/set-by-idp", "/set-by-idp"),
    ],
)
def test_acs(
    request,
    idp_initiated,
    db,
    saml_settings,
    acs_saml_response_factory,
    mocked_signature_validation,
    expected_url,
    relay_state,
    state_kwargs,
    sociallogin_setup_state,
    adv_settings,
    settings,
):
    provider_settings = settings.SOCIALACCOUNT_PROVIDERS["saml"]["APPS"][0]["settings"]
    advanced = dict(provider_settings["advanced"])
    advanced.update(adv_settings)
    provider_settings["advanced"] = advanced
    process = state_kwargs.setdefault("process", AuthProcess.LOGIN)
    is_connect = process == AuthProcess.CONNECT
    if is_connect:
        client = request.getfixturevalue("auth_client")
        user = request.getfixturevalue("user")
    else:
        client = request.getfixturevalue("client")
        user = None

    state_id = None
    if not idp_initiated:
        state_id = sociallogin_setup_state(client, **state_kwargs)

    data = {"SAMLResponse": acs_saml_response_factory(in_response_to=state_id)}
    if relay_state is not None:
        data["RelayState"] = relay_state
    resp = client.post(
        reverse("saml_acs", kwargs={"organization_slug": "org"}), data=data
    )
    finish_url = reverse("saml_finish_acs", kwargs={"organization_slug": "org"})
    assert resp.status_code == 302
    assert resp["location"] == finish_url
    resp = client.get(finish_url)
    if idp_initiated and advanced.get("reject_idp_initiated_sso", True):
        assert "socialaccount/authentication_error.html" in (
            t.name for t in resp.templates
        )
    else:
        assert resp["location"] == expected_url
        account = SocialAccount.objects.get(
            provider="urn:dev-123.us.auth0.com", uid="dummysamluid"
        )
        assert account.extra_data["Role"] == ["view-profile", "manage-account-links"]
        email = EmailAddress.objects.get(user=account.user)
        assert email.email == (user.email if is_connect else "john.doe@email.org")


def test_acs_error(client, db, saml_settings):
    data = {"SAMLResponse": "bad-response"}
    resp = client.post(
        reverse("saml_acs", kwargs={"organization_slug": "org"}), data=data
    )
    assert resp.status_code == 302
    resp = client.get(resp["location"])
    assert "socialaccount/authentication_error.html" in (t.name for t in resp.templates)


def test_acs_get(client, db, saml_settings):
    """WHile ACS expects POST, it always redirects and handles the request in
    the FinishACSView.
    """
    resp = client.get(reverse("saml_acs", kwargs={"organization_slug": "org"}))
    assert resp.status_code == 302
    resp = client.get(resp["location"])
    assert "socialaccount/authentication_error.html" in (t.name for t in resp.templates)


def test_sls_get(client, db, saml_settings):
    """SLS expects POST"""
    resp = client.get(reverse("saml_sls", kwargs={"organization_slug": "org"}))
    assert resp.status_code == 400


def test_login_on_get(client, db, saml_settings):
    resp = client.get(reverse("saml_login", kwargs={"organization_slug": "org"}))
    assert resp.status_code == 200
    assertTemplateUsed(resp, "socialaccount/login.html")


def test_login(client, db, saml_settings):
    resp = client.post(
        reverse("saml_login", kwargs={"organization_slug": "org"})
        + "?process=connect&next=/foo"
    )
    assert resp.status_code == 302
    location = resp["location"]
    assert location.startswith("https://dev-123.us.auth0.com/samlp/456?SAMLRequest=")
    resp_query = parse_qs(urlparse(location).query)
    # We're not using RelayState
    assert resp_query.get("RelayState") is None
    # We're using the request ID / InResponseTo for tracking state.
    state_id = list(client.session[statekit.STATES_SESSION_KEY].keys())[0]
    assert state_id.startswith("ONELOGIN_")
    state = client.session[statekit.STATES_SESSION_KEY][state_id][0]
    assert state == {"process": "connect", "data": None, "next": "/foo"}


def test_metadata(
    client,
    db,
    saml_settings,
):
    resp = client.get(reverse("saml_metadata", kwargs={"organization_slug": "org"}))
    assert resp.status_code == 200
    assert resp.content.startswith(
        b'<?xml version="1.0"?>\n<md:EntityDescriptor xmlns:md="urn:oasis:names:tc:SAML:2.0:metadata'
    )


def test_sls(auth_client, db, saml_settings, user_factory, sls_saml_request):
    with patch("allauth.account.adapter.DefaultAccountAdapter.logout") as logout_mock:
        resp = auth_client.get(
            reverse("saml_sls", kwargs={"organization_slug": "org"})
            + "?"
            + urlencode({"SAMLRequest": sls_saml_request})
        )
        assert logout_mock.call_count == 1
    assert resp.status_code == 302
    assert resp["location"].startswith(
        "https://dev-123.us.auth0.com/samlp/456?SAMLResponse="
    )


@pytest.mark.parametrize(
    "provider_config",
    [
        {
            "idp": {
                "entity_id": "dummy",
                "sso_url": "https://idp.org/sso/",
                "slo_url": "https://idp.saml.org/slo/",
                "x509cert": "cert",
            }
        },
    ],
)
def test_build_saml_config_without_metadata_url(rf, provider_config):
    request = rf.get("/")
    config = build_saml_config(request, provider_config, "org")
    assert config["idp"]["entityId"] == "dummy"
    assert config["idp"]["x509cert"] == "cert"
    assert config["idp"]["singleSignOnService"] == {"url": "https://idp.org/sso/"}
    assert config["idp"]["singleLogoutService"] == {"url": "https://idp.saml.org/slo/"}


@pytest.mark.parametrize(
    "provider_config",
    [
        {
            "idp": {
                "entity_id": "dummy",
                "metadata_url": "https://idp.org/sso/",
            }
        },
        {
            "idp": {
                "entity_id": "dummy",
                "metadata_url": "https://idp.org/sso/",
            },
            "sp": {"entity_id": "dummy-sp-entity-id"},
        },
    ],
)
def test_build_saml_config(rf, provider_config):
    request = rf.get("/")
    with patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.parse_remote"
    ) as parse_mock:
        parse_mock.return_value = {
            "idp": {
                "entityId": "dummy",
                "singleSignOnService": {"url": "https://idp.org/sso/"},
                "singleLogoutService": {"url": "https://idp.saml.org/slo/"},
                "x509cert": "cert",
            }
        }
        config = build_saml_config(request, provider_config, "org")

    assert config["idp"]["entityId"] == "dummy"
    assert config["idp"]["x509cert"] == "cert"
    assert config["idp"]["singleSignOnService"] == {"url": "https://idp.org/sso/"}
    assert config["idp"]["singleLogoutService"] == {"url": "https://idp.saml.org/slo/"}
    metadata_url = request.build_absolute_uri(reverse("saml_metadata", args=["org"]))
    sp_entity_id = provider_config.get("sp", {}).get("entity_id")
    if sp_entity_id:
        assert config["sp"]["entityId"] == sp_entity_id
    else:
        assert config["sp"]["entityId"] == metadata_url


@pytest.mark.parametrize(
    "data, result, uid",
    [
        (
            {"urn:oasis:names:tc:SAML:attribute:subject-id": ["123"]},
            {"uid": "123", "email": "nameid@saml.org"},
            "123",
        ),
        ({}, {"email": "nameid@saml.org"}, "nameid@saml.org"),
    ],
)
def test_extract_attributes(db, data, result, uid, settings):
    settings.SOCIALACCOUNT_PROVIDERS = {
        "saml": {
            "APPS": [
                {
                    "client_id": "org",
                    "provider_id": "urn:dev-123.us.auth0.com",
                }
            ]
        }
    }
    provider = get_adapter().get_provider(request=None, provider="saml")
    onelogin_data = Mock()
    onelogin_data.get_attributes.return_value = data
    onelogin_data.get_nameid.return_value = "nameid@saml.org"
    onelogin_data.get_nameid_format.return_value = (
        "urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
    )
    assert provider._extract(onelogin_data) == result
    assert provider.extract_uid(onelogin_data) == uid



from unittest.mock import Mock

import pytest

from allauth.account.models import EmailAddress
from allauth.socialaccount.providers.base.provider import Provider


@pytest.mark.parametrize(
    "email,emails,expected_email,expected_emails",
    [
        (
            "a@A.COM",
            [EmailAddress(email="A@a.com")],
            "a@a.com",
            [EmailAddress(email="a@a.com")],
        ),
        (
            None,
            [
                EmailAddress(email="A@a.com", primary=True),
                EmailAddress(email="b-AT-b.com"),
                EmailAddress(email="c@c.com", primary=False),
            ],
            "a@a.com",
            [EmailAddress(email="a@a.com"), EmailAddress(email="c@c.com")],
        ),
        (
            "a@A.COM",
            [EmailAddress(email="invalid.com")],
            "a@a.com",
            [EmailAddress(email="a@a.com")],
        ),
        (
            "a@A.COM",
            [EmailAddress(email="is@valid.com")],
            "a@a.com",
            [
                EmailAddress(email="a@a.com"),
                EmailAddress(email="is@valid.com"),
            ],
        ),
        (
            "invalid@test.email",
            [
                EmailAddress(email="invalid@test.email"),
            ],
            None,
            [],
        ),
        (
            "invalid@test.email",
            [
                EmailAddress(email="invalid@test.email"),
                EmailAddress(email="valid@test.email"),
            ],
            "valid@test.email",
            [EmailAddress(email="valid@test.email")],
        ),
    ],
)
def test_cleanup_email_addresses(email, emails, expected_email, expected_emails):
    app = Mock()
    app.settings = {"verified_email": None}
    provider = Provider(None, app=app)
    provider.id = "test"
    email = provider.cleanup_email_addresses(email, emails)
    assert email == expected_email
    assert len(emails) == len(expected_emails)
    for i, addr in enumerate(emails):
        assert addr.email == expected_emails[i].email
